import datetime
import io
from decimal import Decimal
from unittest import TestCase

from a38 import fields, models, validation
from a38.builder import Builder
from a38.diff import Diff

NS = "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.2"


class FieldTestMixin:
    field_class = fields.Field

    def get_field(self, *args, **kw):
        f = self.field_class(*args, **kw)
        f.set_name("sample")
        return f

    def test_xmltag(self):
        # XML tag is autogenerated from the field name
        f = self.get_field()
        self.assertEqual(f.get_xmltag(), "Sample")

        # But can be overridden with the xmltag argument
        f = self.get_field(xmltag="OtherName")
        self.assertEqual(f.get_xmltag(), "OtherName")

    def test_empty(self):
        f = self.get_field()

        # Validating a field with null=False raises an error
        self.assert_validates(f, None, result=None, errors=[
            "sample: missing value",
        ])

        # But null values are tolerated outside validation, while structures
        # are being filled
        self.assertIsNone(f.clean_value(None))

        # Values set to None are skipped in XML
        self.assertIsNone(self.to_xml(f, None))

    def test_nullable(self):
        f = self.get_field(null=True)
        self.assert_validates(f, None, result=None)
        self.assertIsNone(f.clean_value(None))

    def test_construct_default(self):
        f = self.get_field()
        self.assertIsNone(f.get_construct_default())

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, "value", result="value")

    def test_default(self):
        f = self.get_field(default="default")
        self.assertEqual(f.clean_value(None), "default")
        self.assertEqual(self.to_xml(f, None), "<T><Sample>default</Sample></T>")

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertEqual(self.to_xml(f, "value"), "<T><Sample>value</Sample></T>")

    def test_to_python_none(self):
        f = self.get_field()
        self.assert_to_python_works(f, None)

    def test_diff_none(self):
        f = self.get_field()
        self.assert_diff_empty(f, None, None)

    def assert_validates(self, field, value, result, warnings=[], errors=[]):
        val = validation.Validation()
        validated = field.validate(val, value)
        self.assertEqual([str(x) for x in val.warnings], warnings)
        self.assertEqual([str(x) for x in val.errors], errors)
        self.assertEqual(validated, result)

    def assert_to_python_works(self, field, value, **kw):
        kw.setdefault("namespace", False)
        py = field.to_python(value, **kw)
        try:
            parsed = eval(py)
        except Exception as e:
            self.fail("cannot parse generated python {}: {}".format(repr(py), str(e)))
        clean = field.clean_value(parsed)
        self.assertEqual(clean, field.clean_value(value))

    def assert_diff_empty(self, field, first, second):
        """
        Check that the field diff between the two values is empty
        """
        res = Diff()
        field.diff(res, first, second)
        if res.differences:
            self.assertEqual([(d.prefix, d.field, d.first, d.second) for d in res.differences], [])

    def assert_diff(self, field, first, second, expected):
        """
        Check that the field diff between the two differing values, is as
        expected
        """
        res = Diff()
        field.diff(res, first, second)
        self.assertEqual([str(d) for d in res.differences], expected)

    def assert_field_diff(self, field, first, second):
        """
        Check that the field diff, from a non-composite field, between the two
        differing values, is as expected
        """
        self.assert_diff(field, first, None, ["sample: second is not set"])
        self.assert_diff(field, None, first, ["sample: first is not set"])
        self.assert_diff(field, second, None, ["sample: second is not set"])
        self.assert_diff(field, None, second, ["sample: first is not set"])
        self.assert_diff(field, first, second, ["sample: first: {}, second: {}".format(
            field.to_str(field.clean_value(first)),
            field.to_str(field.clean_value(second)))])

    def to_xml(self, field, value):
        """
        Serialize the field to XML. Returns None is the field generated no
        value in the XML.
        """
        builder = Builder()
        with builder.element("T"):
            field.to_xml(builder, value)
        tree = builder.get_tree()
        root = tree.getroot()
        if not list(root):
            return None
        with io.StringIO() as out:
            tree.write(out, encoding="unicode")
            return out.getvalue()

    def mkdt(self, ye, mo, da, ho, mi, se=0, tz=None):
        if tz is None:
            tz = fields.DateTimeField.tz_rome
        return tz.localize(datetime.datetime(ye, mo, da, ho, mi, se))


class TestField(FieldTestMixin, TestCase):
    pass


class TestStringField(FieldTestMixin, TestCase):
    field_class = fields.StringField

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, "value", result="value")
        self.assert_validates(f, 12, result="12")

    def test_default(self):
        f = self.get_field(default="default")
        self.assertEqual(f.clean_value(None), "default")
        self.assertEqual(self.to_xml(f, None), "<T><Sample>default</Sample></T>")

    def test_length(self):
        f = self.get_field(length=3)
        self.assert_validates(f, "va", result="va", errors=[
            "sample: 'va' should be at least 3 characters long",
        ])
        self.assert_validates(f, "valu", result="valu", errors=[
            "sample: 'valu' should be no more than 3 characters long",
        ])
        self.assert_validates(f, 1.15, result="1.15", errors=[
            "sample: '1.15' should be no more than 3 characters long",
        ])
        self.assert_validates(f, "val", result="val")
        self.assert_validates(f, 1.2, result="1.2")

    def test_min_length(self):
        f = self.get_field(min_length=3)
        self.assert_validates(f, "va", result="va", errors=[
            "sample: 'va' should be at least 3 characters long",
        ])
        self.assert_validates(f, "valu", result="valu")
        self.assert_validates(f, "val", result="val")
        self.assert_validates(f, 1.2, result="1.2")
        self.assert_validates(f, 1.15, result="1.15")

    def test_max_length(self):
        f = self.get_field(max_length=3)
        self.assert_validates(f, "v", result="v")
        self.assert_validates(f, "va", result="va")
        self.assert_validates(f, "val", result="val")
        self.assert_validates(f, "valu", result="valu", errors=[
            "sample: 'valu' should be no more than 3 characters long",
        ])

    def test_choices(self):
        f = self.get_field(choices=("A", "B"))
        self.assert_validates(f, "A", result="A")
        self.assert_validates(f, "B", result="B")
        self.assert_validates(f, "C", result="C", errors=[
            "sample: 'C' is not a valid choice for this field",
        ])
        self.assert_validates(f, "a", result="a", errors=[
            "sample: 'a' is not a valid choice for this field",
        ])
        self.assert_validates(f, None, result=None, errors=[
            "sample: missing value",
        ])

    def test_choices_nullable(self):
        f = self.get_field(choices=("A", "B"), null=True)
        self.assert_validates(f, "A", result="A")
        self.assert_validates(f, "B", result="B")
        self.assert_validates(f, None, result=None)
        self.assert_validates(f, "C", result="C", errors=[
            "sample: 'C' is not a valid choice for this field",
        ])
        self.assert_validates(f, "a", result="a", errors=[
            "sample: 'a' is not a valid choice for this field",
        ])

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, "")
        self.assert_to_python_works(f, "foo")
        self.assert_to_python_works(f, "'\"\n")
        self.assert_to_python_works(f, r"\d\t\n")

    def test_diff(self):
        f = self.get_field()
        self.assert_diff_empty(f, "", "")
        self.assert_diff_empty(f, "a", "a")
        self.assert_field_diff(f, "a", "b")

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertEqual(self.to_xml(f, "value"), "<T><Sample>value</Sample></T>")


class TestIntegerField(FieldTestMixin, TestCase):
    field_class = fields.IntegerField

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, 12, result=12)
        self.assert_validates(f, "12", result=12)
        self.assert_validates(f, 12.3, result=12)
        self.assert_validates(f, "foo", result="foo", errors=[
            "sample: invalid literal for int() with base 10: 'foo'",
        ])

    def test_default(self):
        f = self.get_field(default=7)
        self.assertEqual(f.clean_value(None), 7)
        self.assertEqual(self.to_xml(f, None), "<T><Sample>7</Sample></T>")

    def test_max_length(self):
        f = self.get_field(max_length=3)
        self.assert_validates(f, 1, result=1)
        self.assert_validates(f, 12, result=12)
        self.assert_validates(f, 123, result=123)
        self.assert_validates(f, 1234, result=1234, errors=[
            "sample: '1234' should be no more than 3 digits long",
        ])

    def test_choices(self):
        f = self.get_field(choices=(1, 2))
        self.assert_validates(f, 1, result=1)
        self.assert_validates(f, 2, result=2)
        self.assert_validates(f, 3, result=3, errors=[
            "sample: 3 is not a valid choice for this field",
        ])
        self.assert_validates(f, None, result=None, errors=[
            "sample: missing value",
        ])

    def test_choices_nullable(self):
        f = self.get_field(choices=(1, 2), null=True)
        self.assert_validates(f, 1, result=1)
        self.assert_validates(f, 2, result=2)
        self.assert_validates(f, 3, result=3, errors=[
            "sample: 3 is not a valid choice for this field",
        ])
        self.assert_validates(f, None, result=None)

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, 1)
        self.assert_to_python_works(f, 123456)
        self.assert_to_python_works(f, 3 ** 80)

    def test_diff(self):
        f = self.get_field()
        self.assert_diff_empty(f, 1, "1")
        self.assert_field_diff(f, 1, "2")

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertEqual(self.to_xml(f, 1), "<T><Sample>1</Sample></T>")


class TestDecimalField(FieldTestMixin, TestCase):
    field_class = fields.DecimalField

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, 12, result=Decimal("12.00"))
        self.assert_validates(f, "12", result=Decimal("12.00"))
        self.assert_validates(f, "12.345", result=Decimal("12.345"))
        self.assert_validates(f, "foo", result="foo", errors=[
            "sample: 'foo' cannot be converted to Decimal",
        ])

    def test_default(self):
        f = self.get_field(default="7.0")
        self.assertEqual(f.clean_value(None), Decimal("7.0"))
        self.assertEqual(self.to_xml(f, None), "<T><Sample>7.00</Sample></T>")

    def test_max_length(self):
        f = self.get_field(max_length=4)
        self.assert_validates(f, 1, result=Decimal("1.00"))
        # 12 becomes 12.00 which is 5 characters long on a max_length of 4
        self.assert_validates(f, 12, result=Decimal("12.00"), errors=[
            "sample: '12.00' should be no more than 4 digits long",
        ])

    def test_choices(self):
        f = self.get_field(choices=("1.1", "2.2"))
        self.assert_validates(f, "1.1", result=Decimal("1.1"))
        self.assert_validates(f, Decimal("2.2"), result=Decimal("2.2"))
        # 1.1 does not have an exact decimal representation
        self.assert_validates(f, 1.1, result=Decimal("1.100000000000000088817841970012523233890533447265625"), errors=[
            "sample: Decimal('1.100000000000000088817841970012523233890533447265625') is not a valid choice for this field",
        ])
        self.assert_validates(f, None, result=None, errors=[
            "sample: missing value",
        ])

    def test_choices_nullable(self):
        f = self.get_field(choices=("1.1", "2.2"), null=True)
        self.assert_validates(f, "1.1", result=Decimal("1.1"))
        self.assert_validates(f, Decimal("2.2"), result=Decimal("2.2"))
        self.assert_validates(f, None, result=None)
        # 1.1 does not have an exact decimal representation
        dec11 = Decimal(1.1)
        self.assert_validates(f, 1.1, result=dec11, errors=[
            "sample: {!r} is not a valid choice for this field".format(dec11),
        ])

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, "1.2")
        self.assert_to_python_works(f, Decimal("1.20"))

    def test_diff(self):
        f = self.get_field()
        self.assert_diff_empty(f, Decimal("1.0"), "1.0")
        self.assert_diff_empty(f, "1.0001", "1.0002")
        self.assert_field_diff(f, "1.1", "1.2")

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertEqual(self.to_xml(f, "12.345"), "<T><Sample>12.35</Sample></T>")
        self.assertEqual(self.to_xml(f, "34.567"), "<T><Sample>34.57</Sample></T>")


class TestDateField(FieldTestMixin, TestCase):
    field_class = fields.DateField

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, datetime.date(2019, 1, 2), result=datetime.date(2019, 1, 2))
        self.assert_validates(f, "2019-01-02", result=datetime.date(2019, 1, 2))
        self.assert_validates(f, datetime.datetime(2019, 1, 2, 12, 30), result=datetime.date(2019, 1, 2))
        self.assert_validates(f, "foo", result="foo", errors=[
            "sample: Date 'foo' does not begin with YYYY-mm-dd",
        ])
        self.assert_validates(f, [123], result=[123], errors=[
            "sample: '[123]' is not an instance of str, datetime.date or datetime.datetime",
        ])

    def test_default(self):
        f = self.get_field(default="2019-01-02")
        self.assertEqual(f.clean_value(None), datetime.date(2019, 1, 2))
        self.assertEqual(self.to_xml(f, None), "<T><Sample>2019-01-02</Sample></T>")

    def test_choices(self):
        f = self.get_field(choices=("2019-01-01", "2019-01-02"))
        self.assert_validates(f, "2019-01-01", result=datetime.date(2019, 1, 1))
        self.assert_validates(f, "2019-01-02", result=datetime.date(2019, 1, 2))
        self.assert_validates(f, "2019-01-03", result=datetime.date(2019, 1, 3), errors=[
            "sample: datetime.date(2019, 1, 3) is not a valid choice for this field",
        ])
        self.assert_validates(f, None, result=None, errors=[
            "sample: missing value",
        ])

    def test_choices_nullable(self):
        f = self.get_field(choices=("2019-01-01", "2019-01-02"), null=True)
        self.assert_validates(f, "2019-01-01", result=datetime.date(2019, 1, 1))
        self.assert_validates(f, "2019-01-02", result=datetime.date(2019, 1, 2))
        self.assert_validates(f, "2019-01-03", result=datetime.date(2019, 1, 3), errors=[
            "sample: datetime.date(2019, 1, 3) is not a valid choice for this field",
        ])
        self.assert_validates(f, None, result=None)

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, "2019-01-03")
        self.assert_to_python_works(f, datetime.date(2019, 2, 4))

    def test_diff(self):
        f = self.get_field()
        self.assert_diff_empty(f, datetime.date(2019, 1, 1), "2019-01-01")
        self.assert_field_diff(f, datetime.date(2019, 1, 1), "2019-01-02")
        self.assert_field_diff(f, "2019-01-01", "2019-01-02")

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertEqual(self.to_xml(f, datetime.date(2019, 1, 2)), "<T><Sample>2019-01-02</Sample></T>")
        self.assertEqual(self.to_xml(f, "2019-01-02"), "<T><Sample>2019-01-02</Sample></T>")


class TestDateTimeField(FieldTestMixin, TestCase):
    field_class = fields.DateTimeField

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, datetime.datetime(2019, 1, 2, 12, 30), result=self.mkdt(2019, 1, 2, 12, 30))
        self.assert_validates(f, "2019-01-02T12:30:00", result=self.mkdt(2019, 1, 2, 12, 30))
        self.assert_validates(f, datetime.datetime(2019, 1, 2, 12, 30), result=self.mkdt(2019, 1, 2, 12, 30))
        self.assert_validates(f, "foo", result="foo", errors=[
            "sample: ISO string too short",
        ])
        self.assert_validates(f, [123], result=[123], errors=[
            "sample: '[123]' is not an instance of str, datetime.date or datetime.datetime",
        ])

    def test_default(self):
        f = self.get_field(default="2019-01-02T12:30:00")
        self.assertEqual(f.clean_value(None), self.mkdt(2019, 1, 2, 12, 30))
        self.assertEqual(self.to_xml(f, None), "<T><Sample>2019-01-02T12:30:00+01:00</Sample></T>")

    def test_choices(self):
        f = self.get_field(choices=("2019-01-01T12:00:00", "2019-01-02T12:30:00"))
        self.assert_validates(f, "2019-01-01T12:00:00", result=self.mkdt(2019, 1, 1, 12, 00))
        self.assert_validates(f, "2019-01-02T12:30:00", result=self.mkdt(2019, 1, 2, 12, 30))
        self.assert_validates(f, self.mkdt(2019, 1, 2, 12, 15), result=self.mkdt(2019, 1, 2, 12, 15), errors=[
            "sample: 2019-01-02T12:15:00+01:00 is not a valid choice for this field",
        ])
        self.assert_validates(f, None, result=None, errors=[
            "sample: missing value",
        ])

    def test_choices_nullable(self):
        f = self.get_field(choices=("2019-01-01T12:00:00", "2019-01-02T12:30:00"), null=True)
        self.assert_validates(f, "2019-01-01T12:00:00", result=self.mkdt(2019, 1, 1, 12, 00))
        self.assert_validates(f, "2019-01-02T12:30:00", result=self.mkdt(2019, 1, 2, 12, 30))
        self.assert_validates(f, None, result=None)
        self.assert_validates(f, self.mkdt(2019, 1, 2, 12, 15), result=self.mkdt(2019, 1, 2, 12, 15), errors=[
            "sample: 2019-01-02T12:15:00+01:00 is not a valid choice for this field",
        ])

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, "2019-01-03T04:05:06")
        self.assert_to_python_works(f, self.mkdt(2019, 1, 2, 3, 4, 5))

    def test_diff(self):
        f = self.get_field()
        self.assert_diff_empty(f, self.mkdt(2019, 1, 2, 3, 4, 5), "2019-01-02T03:04:05+01:00")
        self.assert_field_diff(f, self.mkdt(2019, 1, 2, 3, 4, 5), self.mkdt(2019, 1, 2, 3, 4, 6))
        self.assert_field_diff(f, self.mkdt(2019, 1, 2, 3, 4, 5), "2019-01-02T03:04:05+02:00")

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertEqual(self.to_xml(f, self.mkdt(2019, 1, 2, 12, 30)), "<T><Sample>2019-01-02T12:30:00+01:00</Sample></T>")
        self.assertEqual(self.to_xml(f, "2019-01-02T12:13:14"), "<T><Sample>2019-01-02T12:13:14+01:00</Sample></T>")


class TestProgressivoInvioField(FieldTestMixin, TestCase):
    field_class = fields.ProgressivoInvioField

    def test_construct_default(self):
        f = self.get_field()

        # The field generates always different, always increasing values
        a = f.get_construct_default()
        b = f.get_construct_default()
        c = f.get_construct_default()
        d = f.get_construct_default()
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertNotEqual(a, d)
        self.assertNotEqual(b, c)
        self.assertNotEqual(b, d)
        self.assertNotEqual(c, d)
        self.assertLess(a, b)
        self.assertLess(b, c)
        self.assertLess(c, d)

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, "BFABFAF")


class Sample(models.Model):
    name = fields.StringField()
    value = fields.IntegerField()


class TestModelField(FieldTestMixin, TestCase):
    field_class = fields.ModelField

    def get_field(self, *args, **kw):
        return super().get_field(Sample, *args, **kw)

    def test_construct_default(self):
        f = self.get_field()
        value = f.get_construct_default()
        self.assertIsInstance(value, Sample)
        self.assertIsNone(value.name)
        self.assertIsNone(value.value)

    def test_empty(self):
        super().test_empty()

        # Empty models are skipped in XML
        f = self.get_field()
        self.assertIsNone(self.to_xml(f, Sample()))

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, Sample("test", 7), result=Sample("test", 7))

    def test_default(self):
        f = self.get_field(default=Sample("test", 7))
        self.assertEqual(f.clean_value(None), Sample("test", 7))
        self.assertEqual(self.to_xml(f, None), "<T><Sample><Name>test</Name><Value>7</Value></Sample></T>")

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, Sample("test", 7))

    def test_diff(self):
        f = self.get_field()
        self.assert_diff_empty(f, Sample("test", 7), Sample("test", "7"))
        self.assert_diff(f, Sample("test", 6), None, [
            "sample: second is not set",
        ])
        self.assert_diff(f, Sample("test", 6), Sample("test", 7), [
            "sample.value: first: 6, second: 7",
        ])
        self.assert_diff(f, Sample("test1", 6), Sample("test2", 7), [
            "sample.name: first: test1, second: test2",
            "sample.value: first: 6, second: 7",
        ])

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertEqual(self.to_xml(f, Sample("test", 7)), "<T><Sample><Name>test</Name><Value>7</Value></Sample></T>")


class TestModelListField(FieldTestMixin, TestCase):
    field_class = fields.ModelListField

    def get_field(self, *args, **kw):
        return super().get_field(Sample, *args, **kw)

    def test_construct_default(self):
        f = self.get_field()
        value = f.get_construct_default()
        self.assertEqual(value, [])

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, [], result=[], errors=[
            "sample: missing value",
        ])
        self.assert_validates(f, [Sample("test", 7)], result=[Sample("test", 7)])

        f = self.get_field(null=True)
        self.assert_validates(f, [], result=[])

    def test_min_num(self):
        f = self.get_field(min_num=2)
        self.assertEqual(f.get_construct_default(), [Sample(), Sample()])
        self.assertEqual(f.clean_value([Sample(), Sample(), Sample()]), [Sample(), Sample()])

        self.assert_validates(f, [Sample("test", 7)], result=[Sample("test", 7)], errors=[
            "sample: list must have at least 2 elements, but has only 1",
        ])
        self.assert_validates(f, [Sample("test", 6), Sample("test", 7)], result=[Sample("test", 6), Sample("test", 7)])

    def test_default(self):
        f = self.get_field(default=[Sample("test", 7)])
        self.assertEqual(f.clean_value(None), [Sample("test", 7)])
        self.assertEqual(self.to_xml(f, None), "<T><Sample><Name>test</Name><Value>7</Value></Sample></T>")

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, [Sample("test", 7)])

    def test_diff(self):
        f = self.get_field()
        self.assert_diff_empty(f, [], None)
        self.assert_diff_empty(f, [Sample("test", 7)], [Sample("test", "7")])
        self.assert_diff_empty(f, [Sample("test", 6)], [Sample("test", 6), None])
        self.assert_diff(f, [Sample("test", 6)], None, [
            "sample: second is not set",
        ])
        self.assert_diff(f, [Sample("test", 6)], [], [
            "sample: second is not set",
        ])
        self.assert_diff(f, [Sample("test", 6)], [Sample("test", 7)], [
            "sample.0.value: first: 6, second: 7",
        ])
        self.assert_diff(f, [Sample("test", 6)], [Sample("test", 6), Sample("test", 7)], [
            "sample: second has 1 extra element",
        ])
        self.assert_diff(f, [Sample("test", 6)], [Sample("test", 5), Sample("test", 7)], [
            "sample.0.value: first: 6, second: 5",
            "sample: second has 1 extra element",
        ])

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertIsNone(self.to_xml(f, []))
        self.assertEqual(self.to_xml(f, [Sample("test", 7)]), "<T><Sample><Name>test</Name><Value>7</Value></Sample></T>")


class TestListField(FieldTestMixin, TestCase):
    field_class = fields.ListField

    def get_field(self, *args, **kw):
        return super().get_field(fields.StringField(), *args, **kw)

    def test_construct_default(self):
        f = self.get_field()
        value = f.get_construct_default()
        self.assertEqual(value, [])

    def test_value(self):
        f = self.get_field()
        self.assert_validates(f, [], result=[], errors=[
            "sample: missing value",
        ])
        self.assert_validates(f, ["test1", "test2"], result=["test1", "test2"])

        f = self.get_field(null=True)
        self.assert_validates(f, [], result=[])

    def test_min_num(self):
        f = self.get_field(min_num=2)
        self.assertEqual(f.get_construct_default(), [None, None])
        self.assertEqual(f.clean_value([None, None, None]), [None, None])

        self.assert_validates(f, ["test1"], result=["test1"], errors=[
            "sample: list must have at least 2 elements, but has only 1",
        ])
        self.assert_validates(f, ["test1", "test2"], result=["test1", "test2"])

    def test_default(self):
        f = self.get_field(default=["test1", "test2"])
        self.assertEqual(f.clean_value(None), ["test1", "test2"])
        self.assertEqual(self.to_xml(f, None), "<T><Sample>test1</Sample><Sample>test2</Sample></T>")

    def test_to_python(self):
        f = self.get_field()
        self.assert_to_python_works(f, ["test1", "foo"])

        f = super().get_field(fields.DateTimeField())
        self.assert_to_python_works(f, [self.mkdt(2019, 1, 2, 3, 4), self.mkdt(2019, 2, 3, 4, 5)])

    def test_diff(self):
        f = self.get_field()
        self.assert_diff_empty(f, [], None)
        self.assert_diff_empty(f, ["test"], ["test"])
        self.assert_diff_empty(f, ["test"], ["test", None])
        self.assert_diff(f, ["test"], ["test1"], [
            "sample.0: first: test, second: test1",
        ])
        self.assert_diff(f, ["test"], ["test", "test"], [
            "sample: second has 1 extra element",
        ])
        self.assert_diff(f, ["test"], ["test1", "test2"], [
            "sample.0: first: test, second: test1",
            "sample: second has 1 extra element",
        ])

    def test_xml(self):
        f = self.get_field(null=True)
        self.assertIsNone(self.to_xml(f, []))
        self.assertEqual(self.to_xml(f, ["test", "foo"]), "<T><Sample>test</Sample><Sample>foo</Sample></T>")



import base64
import binascii
import datetime
import io
import subprocess

try:
    from defusedxml import ElementTree as ET
except ModuleNotFoundError:
    import xml.etree.ElementTree as ET

from typing import BinaryIO, Union

from asn1crypto.cms import ContentInfo

from . import fattura as a38


class SignatureVerificationError(Exception):
    pass


class InvalidSignatureError(SignatureVerificationError):
    pass


class SignerCertificateError(SignatureVerificationError):
    pass


class P7M:
    """
    Parse a Fattura Elettronica encoded as a .p7m file
    """
    def __init__(self, data: Union[str, bytes, BinaryIO]):
        """
        If data is a string, it is taken as a file name.

        If data is bytes, it is taken as p7m data.

        Otherwise, data is taken as a file-like object that reads bytes data.
        """
        if isinstance(data, str):
            with open(data, "rb") as fd:
                self.data = fd.read()
        elif isinstance(data, bytes):
            self.data = data
        else:
            self.data = data.read()

        # Data might potentially be base64 encoded

        try:
            self.data = base64.b64decode(self.data, validate=True)
        except binascii.Error:
            pass

        self.content_info = ContentInfo.load(self.data)

    def is_expired(self) -> bool:
        """
        Check if the signature has expired
        """
        now = datetime.datetime.utcnow()
        signed_data = self.get_signed_data()
        for c in signed_data["certificates"]:
            if c.name != "certificate":
                # The signatures I've seen so far use 'certificate' only
                continue
            expiration_date = c.chosen["tbs_certificate"]["validity"]["not_after"].chosen.native.replace(tzinfo=None)
            if expiration_date <= now:
                return True
        return False

    def get_signed_data(self):
        """
        Return the SignedData part of the P7M file
        """
        if self.content_info["content_type"].native != "signed_data":
            raise RuntimeError("p7m data is not an instance of signed_data")

        signed_data = self.content_info["content"]
        if signed_data["version"].native != "v1":
            raise RuntimeError(f"ContentInfo/SignedData.version is {signed_data['version'].native} instead of v1")

        return signed_data

    def get_payload(self):
        """
        Return the raw signed data
        """
        signed_data = self.get_signed_data()
        encap_content_info = signed_data["encap_content_info"]
        return encap_content_info["content"].native

    def get_fattura(self):
        """
        Return the parsed XML data
        """
        data = io.BytesIO(self.get_payload())
        tree = ET.parse(data)
        return a38.auto_from_etree(tree.getroot())

    def verify_signature(self, certdir):
        """
        Verify the signature on the file
        """
        res = subprocess.run([
            "openssl", "cms", "-verify", "-inform", "DER", "-CApath", certdir, "-noout"],
            input=self.data,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE)

        # From openssl cms manpage:
        # 0   The operation was completely successfully.
        # 1   An error occurred parsing the command options.
        # 2   One of the input files could not be read.
        # 3   An error occurred creating the CMS file or when reading the MIME message.
        # 4   An error occurred decrypting or verifying the message.
        # 5   The message was verified correctly but an error occurred writing out the signers certificates.

        if res.returncode == 0:
            pass
        elif res.returncode == 4:
            raise InvalidSignatureError(res.stderr)
        elif res.returncode == 5:
            raise SignerCertificateError(res.stderr)
        else:
            raise RuntimeError(res.stderr)



from contextlib import contextmanager
from typing import Optional

from . import fields


class Annotation:
    def __init__(self, prefix: Optional[str], field: "fields.Field"):
        self.prefix = prefix
        self.field = field

    @property
    def qualified_field(self) -> str:
        if self.prefix is None:
            return self.field.name
        elif self.field.name is None:
            return self.prefix
        else:
            return self.prefix + "." + self.field.name


class Traversal:
    def __init__(self, prefix: Optional[str] = None):
        self.prefix = prefix

    def with_prefix(self, prefix: str) -> "Traversal":
        raise NotImplementedError("Traversal subclasses must implement with_prefix")

    @contextmanager
    def subfield(self, name: str):
        if self.prefix is None:
            prefix = name
        else:
            prefix = self.prefix + "." + name
        yield self.with_prefix(prefix)



# This module builds XML trees but does not parse them, so it does not need
# defusedxml
import xml.etree.ElementTree as ET
from contextlib import contextmanager

try:
    import lxml.etree
    HAVE_LXML = True
except ModuleNotFoundError:
    HAVE_LXML = False


class Builder:
    def __init__(self, etreebuilder=None):
        if etreebuilder is None:
            etreebuilder = ET.TreeBuilder()
        self.etreebuilder = etreebuilder
        self.default_namespace = None

    def _decorate_tag_name(self, tag: str):
        if self.default_namespace is not None and not tag.startswith("{"):
            return "{" + self.default_namespace + "}" + tag
        return tag

    def add(self, tag: str, value: str, **attrs):
        tag = self._decorate_tag_name(tag)
        self.etreebuilder.start(tag, attrs)
        if value is not None:
            self.etreebuilder.data(value)
        self.etreebuilder.end(tag)

    @contextmanager
    def element(self, tag: str, **attrs):
        tag = self._decorate_tag_name(tag)
        self.etreebuilder.start(tag, attrs)
        yield self
        self.etreebuilder.end(tag)

    @contextmanager
    def override_default_namespace(self, ns):
        b = Builder(self.etreebuilder)
        b.default_namespace = ns
        yield b

    def get_tree(self):
        root = self.etreebuilder.close()
        return ET.ElementTree(root)


if HAVE_LXML:
    class LXMLBuilder:
        def __init__(self, etreebuilder=None):
            if etreebuilder is None:
                etreebuilder = lxml.etree.TreeBuilder()
            self.etreebuilder = etreebuilder
            self.default_namespace = None

        def _decorate_tag_name(self, tag: str):
            if self.default_namespace is not None and not tag.startswith("{"):
                return "{" + self.default_namespace + "}" + tag
            return tag

        def add(self, tag: str, value: str, **attrs):
            tag = self._decorate_tag_name(tag)
            self.etreebuilder.start(tag, attrs)
            if value is not None:
                self.etreebuilder.data(value)
            self.etreebuilder.end(tag)

        @contextmanager
        def element(self, tag: str, **attrs):
            tag = self._decorate_tag_name(tag)
            self.etreebuilder.start(tag, attrs)
            yield self
            self.etreebuilder.end(tag)

        @contextmanager
        def override_default_namespace(self, ns):
            b = Builder(self.etreebuilder)
            b.default_namespace = ns
            yield b

        def get_tree(self):
            root = self.etreebuilder.close()
            return lxml.etree.ElementTree(root)



import base64
import logging
import re
import subprocess

try:
    from defusedxml import ElementTree as ET
except ModuleNotFoundError:
    import xml.etree.ElementTree as ET

from collections import defaultdict
from pathlib import Path
from typing import Dict

from cryptography import x509

from . import fields, models

log = logging.getLogger("__name__")

NS = "http://uri.etsi.org/02231/v2#"
NS_XMLDSIG = "http://www.w3.org/2000/09/xmldsig#"
NS_ADDTYPES = "http://uri.etsi.org/02231/v2/additionaltypes#"


class OtherInformation(models.Model):
    __xmlns__ = NS
    tsl_type = fields.NotImplementedField(xmltag="TSLType", xmlns=NS)
    scheme_territory = fields.StringField(null=True, xmlns=NS)
    mime_type = fields.StringField(null=True, xmlns=NS_ADDTYPES)
    scheme_operator_name = fields.NotImplementedField(xmlns=NS)
    scheme_type_community_rules = fields.NotImplementedField(xmlns=NS)


class AdditionalInformation(models.Model):
    __xmlns__ = NS
    other_information = fields.ModelListField(OtherInformation)


class OtherTSLPointer(models.Model):
    __xmlns__ = NS
    tsl_location = fields.StringField(xmltag="TSLLocation", xmlns=NS)
    service_digital_identities = fields.NotImplementedField(xmlns=NS)
    additional_information = AdditionalInformation


class PointersToOtherTSL(models.Model):
    __xmlns__ = NS
    other_tsl_pointer = fields.ModelListField(OtherTSLPointer)


class SchemeInformation(models.Model):
    __xmlns__ = NS
    pointers_to_other_tsl = fields.ModelField(PointersToOtherTSL)
    tsl_version_identifier = fields.NotImplementedField(xmltag="TSLVersionIdentifier", xmlns=NS)
    tsl_sequence_number = fields.NotImplementedField(xmltag="TSLSequenceNumber", xmlns=NS)
    tsl_type = fields.NotImplementedField(xmltag="TSLType", xmlns=NS)
    scheme_operator_name = fields.NotImplementedField(xmlns=NS)
    scheme_operator_address = fields.NotImplementedField(xmlns=NS)
    scheme_information_uri = fields.NotImplementedField(xmltag="SchemeInformationURI", xmlns=NS)
    scheme_name = fields.NotImplementedField(xmlns=NS)
    status_determination_approach = fields.NotImplementedField(xmlns=NS)
    scheme_type_community_rules = fields.NotImplementedField(xmlns=NS)
    scheme_territory = fields.NotImplementedField(xmlns=NS)
    policy_or_legal_notice = fields.NotImplementedField(xmlns=NS)
    historical_information_period = fields.NotImplementedField(xmlns=NS)
    list_issue_date_time = fields.NotImplementedField(xmlns=NS)
    next_update = fields.NotImplementedField(xmlns=NS)
    distribution_points = fields.NotImplementedField(xmlns=NS)


class TSPInformation(models.Model):
    __xmlns__ = NS
    tsp_name = fields.NotImplementedField(xmltag="TSPName", xmlns=NS)
    tsp_trade_name = fields.NotImplementedField(xmltag="TSPTradeName", xmlns=NS)
    tsp_address = fields.NotImplementedField(xmltag="TSPAddress", xmlns=NS)
    tsp_information_url = fields.NotImplementedField(xmltag="TSPInformationURI", xmlns=NS)


class DigitalId(models.Model):
    __xmlns__ = NS
    x509_subject_name = fields.StringField(xmltag="X509SubjectName", xmlns=NS, null=True)
    x509_ski = fields.StringField(xmltag="X509SKI", xmlns=NS, null=True)
    x509_certificate = fields.StringField(xmltag="X509Certificate", xmlns=NS, null=True)


class ServiceDigitalIdentity(models.Model):
    __xmlns__ = NS
    digital_id = fields.ModelListField(DigitalId)


class ServiceInformation(models.Model):
    __xmlns__ = NS
    service_type_identifier = fields.StringField(xmlns=NS)
    service_name = fields.NotImplementedField(xmlns=NS)
    service_digital_identity = ServiceDigitalIdentity
    service_status = fields.StringField(xmlns=NS)
    status_starting_time = fields.NotImplementedField(xmlns=NS)
    service_information_extensions = fields.NotImplementedField(xmlns=NS)


class TSPService(models.Model):
    __xmlns__ = NS
    service_information = ServiceInformation
    service_history = fields.NotImplementedField(xmlns=NS)


class TSPServices(models.Model):
    __xmlns__ = NS
    tsp_service = fields.ModelListField(TSPService)


class TrustServiceProvider(models.Model):
    __xmlns__ = NS
    tsp_information = TSPInformation
    tsp_services = TSPServices


class TrustServiceProviderList(models.Model):
    __xmlns__ = NS
    trust_service_provider = fields.ModelListField(TrustServiceProvider)


class TrustServiceStatusList(models.Model):
    __xmlns__ = NS
    scheme_information = SchemeInformation
    signature = fields.NotImplementedField(xmlns=NS_XMLDSIG)
    trust_service_provider_list = TrustServiceProviderList

    def get_tsl_pointer_by_territory(self, territory):
        for other_tsl_pointer in self.scheme_information.pointers_to_other_tsl.other_tsl_pointer:
            territory = None
            for oi in other_tsl_pointer.additional_information.other_information:
                if oi.scheme_territory is not None:
                    territory = oi.scheme_territory
                    break
            if territory != "IT":
                continue
            return other_tsl_pointer.tsl_location


def auto_from_etree(root):
    expected_tag = "{{{}}}TrustServiceStatusList".format(NS)
    if root.tag != expected_tag:
        raise RuntimeError("Root element {} is not {}".format(root.tag, expected_tag))

    res = TrustServiceStatusList()
    res.from_etree(root)
    return res


def load_url(url: str):
    """
    Return a TrustedServiceStatusList instance from the XML downloaded from the
    given URL
    """
    import requests
    res = requests.get(url)
    res.raise_for_status()
    root = ET.fromstring(res.content)
    return auto_from_etree(root)


def load_certs() -> Dict[str, x509.Certificate]:
    """
    Download trusted list certificates for Italy, parse them and return a dict
    mapping certificate names good for use as file names to cryptography.x509
    certificates
    """
    re_clean_fname = re.compile(r"[^A-Za-z0-9_-]")

    eu_url = "https://ec.europa.eu/information_society/policy/esignature/trusted-list/tl-mp.xml"
    log.info("Downloading EU index from %s", eu_url)
    eu_tl = load_url(eu_url)
    it_url = eu_tl.get_tsl_pointer_by_territory("IT")
    log.info("Downloading IT data from %s", it_url)
    trust_service_status_list = load_url(it_url)

    by_name = defaultdict(list)
    for tsp in trust_service_status_list.trust_service_provider_list.trust_service_provider:
        for tsp_service in tsp.tsp_services.tsp_service:
            si = tsp_service.service_information
            if si.service_status not in (
                    "http://uri.etsi.org/TrstSvc/TrustedList/Svcstatus/recognisedatnationallevel",
                    "http://uri.etsi.org/TrstSvc/TrustedList/Svcstatus/granted"):
                continue
            if si.service_type_identifier not in (
                    "http://uri.etsi.org/TrstSvc/Svctype/CA/QC",):
                continue
            # print("identifier", si.service_type_identifier)
            # print("status", si.service_status)
            cert = []
            sn = []
            for di in si.service_digital_identity.digital_id:
                if di.x509_subject_name is not None:
                    sn.append(di.x509_subject_name)
                # if di.x509_ski is not None:
                #    print("  SKI:", di.x509_ski)
                if di.x509_certificate is not None:
                    from cryptography import x509
                    from cryptography.hazmat.backends import default_backend
                    der = base64.b64decode(di.x509_certificate)
                    cert.append(x509.load_der_x509_certificate(der, default_backend()))

            if len(cert) == 0:
                raise RuntimeError("{} has no certificates".format(sn))
            elif len(cert) > 1:
                raise RuntimeError("{} has {} certificates".format(sn, len(cert)))
            else:
                from cryptography.x509.oid import NameOID
                cert = cert[0]
                cn = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
                # print("sn", sn)
                # print(cert)
                # print("full cn", cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME))
                # print("cn", cn)
                fname = re_clean_fname.sub("_", cn)
                by_name[fname].append(cert)

    res = {}
    for name, certs in by_name.items():
        if len(certs) == 1:
            if name in res:
                raise RuntimeError("{} already in result".format(name))
            res[name] = certs[0]
        else:
            for idx, cert in enumerate(certs, start=1):
                idxname = name + "_a38_{}".format(idx)
                if idxname in res:
                    raise RuntimeError("{} already in result".format(name))
                res[idxname] = cert
    return res


def update_capath(destdir: Path, remove_old=False):
    from cryptography.hazmat.primitives import serialization
    certs = load_certs()
    if destdir.is_dir():
        current = set(c.name for c in destdir.iterdir() if c.name.endswith(".crt"))
    else:
        current = set()
        destdir.mkdir(parents=True)
    for name, cert in certs.items():
        fname = name + ".crt"
        current.discard(fname)
        pathname = destdir / fname
        with pathname.open(mode="wb") as fd:
            fd.write(cert.public_bytes(serialization.Encoding.PEM))
            log.info("%s: written", pathname)
    if remove_old:
        for fname in current:
            pathname = destdir / fname
            pathname.unlink()
            log.info("%s: removed", pathname)

    subprocess.run(["openssl", "rehash", destdir], check=True)



from __future__ import annotations

from . import consts, fields, models
from .fattura import (Allegati, FullNameMixin, IdFiscaleIVA, IdTrasmittente,
                      IscrizioneREA, Sede, StabileOrganizzazione)

NS10 = "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.0"


class DatiTrasmissione(models.Model):
    id_trasmittente = IdTrasmittente
    progressivo_invio = fields.ProgressivoInvioField()
    formato_trasmissione = fields.StringField(length=5, choices=("FSM10",))
    codice_destinatario = fields.StringField(null=True, min_length=6, max_length=7, default="0000000")
    pec_destinatario = fields.StringField(null=True, min_length=8, max_length=256, xmltag="PECDestinatario")


class RappresentanteFiscale(FullNameMixin, models.Model):
    id_fiscale_iva = IdFiscaleIVA
    denominazione = fields.StringField(max_length=80, null=True)
    nome = fields.StringField(max_length=60, null=True)
    cognome = fields.StringField(max_length=60, null=True)


class CedentePrestatore(FullNameMixin, models.Model):
    id_fiscale_iva = IdFiscaleIVA
    codice_fiscale = fields.StringField(min_length=11, max_length=16, null=True)
    denominazione = fields.StringField(max_length=80, null=True)
    nome = fields.StringField(max_length=60, null=True)
    cognome = fields.StringField(max_length=60, null=True)
    sede = Sede
    stabile_organizzazione = fields.ModelField(StabileOrganizzazione, null=True)
    rappresentante_fiscale = models.ModelField(RappresentanteFiscale, null=True)
    iscrizione_rea = fields.ModelField(IscrizioneREA, null=True)
    regime_fiscale = fields.StringField(
            length=4, choices=("RF01", "RF02", "RF04", "RF05", "RF06", "RF07",
                               "RF08", "RF09", "RF10", "RF11", "RF12", "RF13",
                               "RF14", "RF15", "RF16", "RF17", "RF18", "RF19"))


class IdentificativiFiscali(models.Model):
    id_fiscale_iva = IdFiscaleIVA
    codice_fiscale = fields.StringField(min_length=11, max_length=16, null=True)


class AltriDatiIdentificativi(FullNameMixin, models.Model):
    denominazione = fields.StringField(max_length=80, null=True)
    nome = fields.StringField(max_length=60, null=True)
    cognome = fields.StringField(max_length=60, null=True)
    sede = Sede
    stabile_organizzazione = fields.ModelField(StabileOrganizzazione, null=True)
    rappresentante_fiscale = models.ModelField(RappresentanteFiscale, null=True)


class CessionarioCommittente(models.Model):
    identificativi_fiscali = IdentificativiFiscali
    altri_dati_identificativi = AltriDatiIdentificativi


class FatturaElettronicaHeader(models.Model):
    dati_trasmissione = DatiTrasmissione
    cedente_prestatore = CedentePrestatore
    cessionario_committente = CessionarioCommittente
    soggetto_emittente = fields.StringField(length=2, choices=("CC", "TZ"), null=True)


class DatiGeneraliDocumento(models.Model):
    tipo_documento = fields.StringField(length=4, choices=("TD07", "TD08", "TD09"))
    divisa = fields.StringField()
    data = fields.DateField()
    numero = fields.StringField(max_length=20)


class DatiFatturaRettificata(models.Model):
    numero_fr = fields.StringField(max_length=20, xmltag="NumeroFR")
    data_fr = fields.DateField(xmltag="DataFR")
    elementi_rettificati = fields.StringField(max_length=1000)


class DatiGenerali(models.Model):
    dati_generali_documento = DatiGeneraliDocumento
    dati_fattura_rettificata = fields.ModelField(DatiFatturaRettificata, null=True)


class DatiIVA(models.Model):
    imposta = fields.DecimalField(max_length=15)
    aliquota = fields.DecimalField(max_length=6)


class DatiBeniServizi(models.Model):
    descrizione = fields.StringField(max_length=1000)
    importo = fields.DecimalField(max_length=15)
    dati_iva = DatiIVA
    natura = fields.StringField(length=2, null=True, choices=consts.NATURA_IVA)
    riferimento_normativo = fields.StringField(max_length=100, null=True)


class FatturaElettronicaBody(models.Model):
    dati_generali = DatiGenerali
    dati_beni_servizi = fields.ModelListField(DatiBeniServizi)
    allegati = fields.ModelListField(Allegati, null=True)


class FatturaElettronicaSemplificata(models.Model):
    """
    Fattura elettronica semplificata
    """
    __xmlns__ = NS10
    fattura_elettronica_header = FatturaElettronicaHeader
    fattura_elettronica_body = fields.ModelListField(FatturaElettronicaBody, min_num=1)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.fattura_elettronica_header.dati_trasmissione.formato_trasmissione = self.get_versione()

    def get_versione(self):
        return "FSM10"

    def get_xmlattrs(self):
        return {"versione": self.get_versione()}

    def validate_model(self, validation):
        super().validate_model(validation)
        if self.get_versione() != self.fattura_elettronica_header.dati_trasmissione.formato_trasmissione:
            validation.add_error(
                    self.fattura_elettronica_header.dati_trasmissione._meta["formato_trasmissione"],
                    "formato_trasmissione should be {}".format(self.get_versione()),
                    code="00428")

    def to_xml(self, builder):
        with builder.element(self.get_xmltag(), **self.get_xmlattrs()) as b:
            with b.override_default_namespace(None) as b1:
                for name, field in self._meta.items():
                    field.to_xml(b1, getattr(self, name))

    def build_etree(self, lxml=False):
        """
        Build and return an ElementTree with the fattura in XML format
        """
        self.fattura_elettronica_header.dati_trasmissione.formato_trasmissione = self.get_versione()
        if lxml:
            from a38.builder import LXMLBuilder
            builder = LXMLBuilder()
        else:
            from a38.builder import Builder
            builder = Builder()
        builder.default_namespace = NS10
        self.to_xml(builder)
        return builder.get_tree()

    def from_etree(self, el):
        versione = el.attrib.get("versione", None)
        if versione is None:
            raise RuntimeError("root element {} misses attribute 'versione'".format(el.tag))

        if versione != self.get_versione():
            raise RuntimeError("root element versione is {} instead of {}".format(versione, self.get_versione()))

        return super().from_etree(el)



import os
import tempfile
from contextlib import contextmanager
from unittest import TestCase

from a38.crypto import P7M, InvalidSignatureError

# This is the CA certificate used to validate tests/data/test.txt.p7m
#
# The signature on the test file will expire (next expiration date: May  6 23:59:59 2024 GMT)
#
# To refresh it:
#
# 1. Sign tests/data/test.txt with a CAdES envelope
# 2. Extract the signature:
#      openssl smime -verify -in tests/data/test.txt.p7m -inform der -noverify -signer /tmp/cert.pem -out /dev/null
# 3. Get signature information:
#      openssl x509 -inform pem -in /tmp/cert.pem -text
# 4. Compute the issuer hash to find the CA certificate:
#      openssl x509 -inform pem -in /tmp/cert.pem -issuer_hash
# 5. Download/refresh the CA certificate database:
#      ./a38tool update_capath certs
# 6. Find the file named with the issuer hash in certs/
# 7. Update the CA_CERT_HASH variable below with the name of the file you just
#    found in certs/
# 8. Replace the value of CA_CERT with its contents
#
CA_CERT = """
-----BEGIN CERTIFICATE-----
MIIE+jCCA+KgAwIBAgIQbK2AXjA4PMWG8x+rL26V9zANBgkqhkiG9w0BAQsFADBs
MQswCQYDVQQGEwJJVDEYMBYGA1UECgwPQXJ1YmFQRUMgUy5wLkEuMSEwHwYDVQQL
DBhDZXJ0aWZpY2F0aW9uIEF1dGhvcml0eUMxIDAeBgNVBAMMF0FydWJhUEVDIFMu
cC5BLiBORyBDQSAzMB4XDTEwMTAyMjAwMDAwMFoXDTMwMTAyMjIzNTk1OVowbDEL
MAkGA1UEBhMCSVQxGDAWBgNVBAoMD0FydWJhUEVDIFMucC5BLjEhMB8GA1UECwwY
Q2VydGlmaWNhdGlvbiBBdXRob3JpdHlDMSAwHgYDVQQDDBdBcnViYVBFQyBTLnAu
QS4gTkcgQ0EgMzCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKtkY4EH
G+Nh4VYLL4R5tvmX6J+AYlL2BPDUCLN92+zi9QMbsh84zbRE+om9KE8P67mST2my
bhGTz6dzeK1BrQfSdKJ8AGxePzqUq+uGHGULoy4A6ey4EyqTfxY+pGzjB7OVcuiw
y7iV6k1YjshIsmNjTmYOAQepZMgBmxHPnR6IW9MsAOFBBQH/vJFQDeBts/rA6lbM
/VsURwzr6XOqCzwJK/csKvuE/rAaRKY+IPzah8mou//yEi4V401J1JYfPanbCJOW
nIty9HaioUe5Fu2jw4UP7T5Cbw4lND1sP7HVhsVRDuTj3gF9ulJ7EBmcR/2THDZC
ozD76uwuTmkm4VsCAwEAAaOCAZYwggGSMD8GCCsGAQUFBwEBBDMwMTAvBggrBgEF
BQcwAYYjaHR0cDovL29jc3AuYXJ1YmFwZWMudHJ1c3RpdGFsaWEuaXQwEgYDVR0T
AQH/BAgwBgEB/wIBADBGBgNVHSAEPzA9MDsGCisGAQQBgegtAQEwLTArBggrBgEF
BQcCARYfaHR0cHM6Ly9jYS5hcnViYXBlYy5pdC9jcHMuaHRtbDBqBgNVHR8EYzBh
MF+gXaBbhllodHRwOi8vb25zaXRlY3JsLmFydWJhcGVjLnRydXN0aXRhbGlhLml0
L0FydWJhUEVDU3BBQ2VydGlmaWNhdGlvbkF1dGhvcml0eUMvTGF0ZXN0Q1JMLmNy
bDArBgNVHRIEJDAipCAwHjEcMBoGA1UEAxMTR09WVlNQLUMxLTIwNDgtMS0xMDAO
BgNVHQ8BAf8EBAMCAQYwKwYDVR0RBCQwIqQgMB4xHDAaBgNVBAMTE0dPVlZTUC1D
MS0yMDQ4LTEtMTAwHQYDVR0OBBYEFPDARbG2NbTqXyn6gwNK3C/1s33oMA0GCSqG
SIb3DQEBCwUAA4IBAQBRGwGypquxMawPV6ZN5l/2eJdaaqgnYolin1PGXJUFRQy3
k5FK0Fwk/90U/j/ue83cYdsRpPVpo17LOk7hCNSFk/W2SRVGvqaM77/cVpgFwm25
Ab2x5sMxwJ9Uoouba00CDl2SiYgn9KN+Bd3LHrwtpO8IkzwSE7k0kKmDLdCZTyUO
ZPR8RKpwedjLJoiyXCtq9PKA3avI1R6N8yOxbK954+nSOsHfmGDP4wQi8PUJIWBm
dlpHNM669BLdLwj6lpCjNI6AuP4K5Jw1qkOmcccnVWxkk0r2qNu87AlVosHpKf6G
jkJbJNWfBsgjRHGg6Pq3enAf8/7DfkoCyKUzI8zZ
-----END CERTIFICATE-----
"""

CA_CERT_HASH = "b72ed47c.0"


class TestSignature(TestCase):
    @contextmanager
    def capath(self):
        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, CA_CERT_HASH), "wt") as fd:
                fd.write(CA_CERT)
            yield td

    def test_load(self):
        p7m = P7M("tests/data/test.txt.p7m")
        data = p7m.get_payload()
        self.assertEqual(
                data,
                "This is only a test payload.\n"
                "\n"
                "Questo è solo un payload di test.\n".encode("utf8"))

    def test_verify(self):
        p7m = P7M("tests/data/test.txt.p7m")
        if p7m.is_expired():
            self.skipTest("test signature has expired and needs to be regenerated")
        with self.capath() as capath:
            p7m.verify_signature(capath)

    def test_verify_corrupted_random(self):
        p7m = P7M("tests/data/test.txt.p7m")
        if p7m.is_expired():
            self.skipTest("test signature has expired and needs to be regenerated")
        data_mid = len(p7m.data) // 2
        p7m.data = p7m.data[:data_mid] + bytes([p7m.data[data_mid] + 1]) + p7m.data[data_mid + 1:]
        with self.capath() as capath:
            with self.assertRaises(InvalidSignatureError):
                p7m.verify_signature(capath)

    def test_verify_corrupted_payload(self):
        p7m = P7M("tests/data/test.txt.p7m")
        if p7m.is_expired():
            self.skipTest("test signature has expired and needs to be regenerated")
        signed_data = p7m.get_signed_data()
        encap_content_info = signed_data["encap_content_info"]
        encap_content_info["content"] = b"All your base are belong to us"
        p7m.data = p7m.content_info.dump()
        with self.capath() as capath:
            with self.assertRaisesRegexp(InvalidSignatureError, r"routines:CMS_verify:content verify error"):
                p7m.verify_signature(capath)

    def test_verify_noca(self):
        p7m = P7M("tests/data/test.txt.p7m")
        if p7m.is_expired():
            self.skipTest("test signature has expired and needs to be regenerated")
        with tempfile.TemporaryDirectory() as capath:
            with self.assertRaisesRegexp(
                    InvalidSignatureError, r"Verify error:\s*unable to get local issuer certificate"):
                p7m.verify_signature(capath)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Dict, Union

from . import consts, fields, models

if TYPE_CHECKING:
    import xml.etree.ElementTree as ET

    from .fattura_semplificata import FatturaElettronicaSemplificata

#
# This file describes the data model of the Italian Fattura Elettronica.
#
# Models and fields are inspired from Django's ORM.
#
# XML tag names are built automatically from the field names, and can be
# specified explicitly with the xmltag argument.
#
# Models can be used as fields using `fields.ModelField`. Specifying a Model
# class as a field, automatically wraps it in a `ModelField`.
#

__all__ = []


def export(el):
    """
    Add a symbol to __all__ for export
    """
    global __all__
    __all__.append(el.__name__)
    return el


NS = "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.2"
NS10 = "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.0"
NS_SIG = "http://www.w3.org/2000/09/xmldsig#"


class FullNameMixin:
    """
    Helper for classes that have the nome+cognome/denominazione way of naming.

    Validate that nome+cognome and denominazione are mutually exclusive, and
    provide a full_name property that returns whichever is set.
    """

    @property
    def full_name(self):
        """
        Return denominazione or "{nome} {cognome}", whichever is set.

        If none are set, return None
        """
        if self.denominazione is not None:
            return self.denominazione
        elif self.nome is not None and self.cognome is not None:
            return self.nome + " " + self.cognome
        else:
            return None

    def validate_model(self, validation):
        super().validate_model(validation)
        if self.denominazione is None:
            if self.nome is None and self.cognome is None:
                validation.add_error(
                        (self._meta["nome"], self._meta["cognome"], self._meta["denominazione"]),
                        "nome and cognome, or denominazione, must be set")
            elif self.nome is None:
                validation.add_error(
                        self._meta["nome"],
                        "nome and cognome must both be set if denominazione is empty")
            elif self.cognome is None:
                validation.add_error(
                        self._meta["cognome"],
                        "nome and cognome must both be set if denominazione is empty")
        else:
            should_not_be_set = []
            if self.nome is not None:
                should_not_be_set.append(self._meta["nome"])
            if self.cognome is not None:
                should_not_be_set.append(self._meta["cognome"])
            if should_not_be_set:
                validation.add_error(
                        should_not_be_set,
                        "{} must not be set if denominazione is not empty".format(
                            " and ".join(x.name for x in should_not_be_set)))


@export
class IdFiscale(models.Model):
    id_paese = fields.StringField(length=2)
    id_codice = fields.StringField(max_length=28)


@export
class IdTrasmittente(IdFiscale):
    pass


@export
class IdFiscaleIVA(IdFiscale):
    pass


@export
class ContattiTrasmittente(models.Model):
    telefono = fields.StringField(min_length=5, max_length=12, null=True)
    email = fields.StringField(min_length=7, max_length=256, null=True)


@export
class DatiTrasmissione(models.Model):
    id_trasmittente = IdTrasmittente
    progressivo_invio = fields.ProgressivoInvioField()
    formato_trasmissione = fields.StringField(length=5, choices=("FPR12", "FPA12"))
    codice_destinatario = fields.StringField(null=True, min_length=6, max_length=7, default="0000000")
    contatti_trasmittente = fields.ModelField(ContattiTrasmittente, null=True)
    pec_destinatario = fields.StringField(null=True, min_length=8, max_length=256, xmltag="PECDestinatario")

    def validate_model(self, validation):
        super().validate_model(validation)
        if self.codice_destinatario is None and self.pec_destinatario is None:
            validation.add_error(
                    (self._meta["codice_destinatario"], self._meta["pec_destinatario"]),
                    "one of codice_destinatario or pec_destinatario must be set")

        # Se la fattura deve essere recapitata ad un soggetto che intende
        # ricevere le fatture elettroniche attraverso il canale PEC, il campo
        # deve essere valorizzato con sette zeri (“0000000”) e deve essere
        # valorizzato il campo PECDestinatario

        if self.pec_destinatario is None and self.codice_destinatario == "0000000":
            validation.add_error(
                    (self._meta["codice_destinatario"], self._meta["pec_destinatario"]),
                    "pec_destinatario has no value while codice_destinatario has value 0000000",
                    code="00426")

        if (self.pec_destinatario is not None and self.codice_destinatario is
                not None and self.codice_destinatario != "0000000"):
            validation.add_error(
                    (self._meta["codice_destinatario"], self._meta["pec_destinatario"]),
                    "pec_destinatario has value while codice_destinatario has value 0000000",
                    code="00426")

        if self.formato_trasmissione == "FPA12" and len(self.codice_destinatario) == 7:
            validation.add_error(
                    self._meta["codice_destinatario"],
                    "codice_destinatario has 7 characters on a Fattura PA",
                    code="00427")

        if self.formato_trasmissione == "FPR12" and len(self.codice_destinatario) == 6:
            validation.add_error(
                    self._meta["codice_destinatario"],
                    "codice_destinatario has 6 characters on a Fattura Privati",
                    code="00427")


@export
class Anagrafica(FullNameMixin, models.Model):
    denominazione = fields.StringField(max_length=80, null=True)
    nome = fields.StringField(max_length=60, null=True)
    cognome = fields.StringField(max_length=60, null=True)
    titolo = fields.StringField(min_length=2, max_length=10, null=True)
    cod_eori = fields.StringField(xmltag="CodEORI", min_length=13, max_length=17, null=True)


@export
class DatiAnagraficiCedentePrestatore(models.Model):
    __xmltag__ = "DatiAnagrafici"
    id_fiscale_iva = IdFiscaleIVA
    codice_fiscale = fields.StringField(min_length=11, max_length=16, null=True)
    anagrafica = Anagrafica
    albo_professionale = fields.StringField(max_length=60, null=True)
    provincia_albo = fields.StringField(length=2, null=True)
    numero_iscrizione_albo = fields.StringField(max_length=60, null=True)
    data_iscrizione_albo = fields.DateField(null=True)
    regime_fiscale = fields.StringField(
            length=4, choices=consts.REGIME_FISCALE)


@export
class IndirizzoType(models.Model):
    indirizzo = fields.StringField(max_length=60)
    numero_civico = fields.StringField(max_length=8, null=True)
    cap = fields.StringField(xmltag="CAP", length=5)
    comune = fields.StringField(max_length=60)
    provincia = fields.StringField(length=2, null=True)
    nazione = fields.StringField(length=2)


@export
class Sede(IndirizzoType):
    pass


@export
class IscrizioneREA(models.Model):
    ufficio = fields.StringField(length=2)
    numero_rea = fields.StringField(xmltag="NumeroREA", max_length=20)
    capitale_sociale = fields.StringField(min_length=4, max_length=15, null=True)
    socio_unico = fields.StringField(length=2, choices=("SU", "SM"), null=True)
    stato_liquidazione = fields.StringField(length=2, choices=("LS", "LN"))


@export
class Contatti(models.Model):
    telefono = fields.StringField(min_length=5, max_length=12, null=True)
    fax = fields.StringField(min_length=5, max_length=12, null=True)
    email = fields.StringField(min_length=7, max_length=256, null=True)


@export
class StabileOrganizzazione(IndirizzoType):
    pass


@export
class CedentePrestatore(models.Model):
    dati_anagrafici = DatiAnagraficiCedentePrestatore
    sede = Sede
    stabile_organizzazione = fields.ModelField(StabileOrganizzazione, null=True)
    iscrizione_rea = fields.ModelField(IscrizioneREA, null=True)
    contatti = fields.ModelField(Contatti, null=True)
    riferimento_amministrazione = fields.StringField(max_length=20, null=True)


@export
class DatiAnagraficiCessionarioCommittente(models.Model):
    __xmltag__ = "DatiAnagrafici"
    id_fiscale_iva = fields.ModelField(IdFiscaleIVA, null=True)
    codice_fiscale = fields.StringField(min_length=11, max_length=16, null=True)
    anagrafica = Anagrafica

    def validate_model(self, validation):
        super().validate_model(validation)
        if self.id_fiscale_iva is None and self.codice_fiscale is None:
            validation.add_error(
                    (self._meta["id_fiscale_iva"], self._meta["codice_fiscale"]),
                    "at least one of id_fiscale_iva and codice_fiscale needs to have a value",
                    code="00417")


@export
class RappresentanteFiscale(FullNameMixin, models.Model):
    id_fiscale_iva = fields.ModelField(IdFiscaleIVA, null=True)
    denominazione = fields.StringField(max_length=80, null=True)
    nome = fields.StringField(max_length=60, null=True)
    cognome = fields.StringField(max_length=60, null=True)


@export
class CessionarioCommittente(models.Model):
    dati_anagrafici = DatiAnagraficiCessionarioCommittente
    sede = Sede
    stabile_organizzazione = fields.ModelField(StabileOrganizzazione, null=True)
    rappresentante_fiscale = fields.ModelField(RappresentanteFiscale, null=True)


@export
class DatiAnagraficiRappresentante(models.Model):
    __xmltag__ = "DatiAnagrafici"
    id_fiscale_iva = IdFiscaleIVA
    codice_fiscale = fields.StringField(min_length=11, max_length=16, null=True)
    anagrafica = Anagrafica


@export
class RappresentanteFiscaleCedentePrestatore(models.Model):
    __xmltag__ = "RappresentanteFiscale"
    dati_anagrafici = DatiAnagraficiRappresentante


@export
class DatiAnagraficiTerzoIntermediario(models.Model):
    __xmltag__ = "DatiAnagrafici"
    id_fiscale_iva = IdFiscaleIVA
    codice_fiscale = fields.StringField(min_length=11, max_length=16, null=True)
    anagrafica = Anagrafica


@export
class TerzoIntermediarioOSoggettoEmittente(models.Model):
    dati_anagrafici = DatiAnagraficiTerzoIntermediario


@export
class FatturaElettronicaHeader(models.Model):
    dati_trasmissione = DatiTrasmissione
    cedente_prestatore = CedentePrestatore
    rappresentante_fiscale = models.ModelField(RappresentanteFiscaleCedentePrestatore, null=True)
    cessionario_committente = CessionarioCommittente
    terzo_intermediario_o_soggetto_emittente = models.ModelField(TerzoIntermediarioOSoggettoEmittente, null=True)
    soggetto_emittente = fields.StringField(length=2, choices=("CC", "TZ"), null=True)


@export
class DatiRitenuta(models.Model):
    tipo_ritenuta = fields.StringField(length=4, choices=consts.TIPO_RITENUTA)
    importo_ritenuta = fields.DecimalField(max_length=15)
    aliquota_ritenuta = fields.DecimalField(max_length=6)
    causale_pagamento = fields.StringField(max_length=2)


@export
class DatiBollo(models.Model):
    bollo_virtuale = fields.StringField(length=2, choices=("SI",))
    importo_bollo = fields.DecimalField(max_length=15)


@export
class DatiCassaPrevidenziale(models.Model):
    tipo_cassa = fields.StringField(length=4, choices=consts.TIPO_CASSA)
    al_cassa = fields.DecimalField(max_length=6)
    importo_contributo_cassa = fields.DecimalField(max_length=15)
    imponibile_cassa = fields.DecimalField(max_length=15)
    aliquota_iva = fields.DecimalField(max_length=6, xmltag="AliquotaIVA")
    ritenuta = fields.StringField(length=2, choices=("SI",), null=True)
    natura = fields.StringField(min_length=2, max_length=4, choices=consts.NATURA_IVA, null=True)
    riferimento_amministrazione = fields.StringField(max_length=20, null=True)

    def validate_model(self, validation):
        super().validate_model(validation)
        if self.aliquota_iva == 0 and self.natura is None:
            validation.add_error(
                self._meta["natura"], "field is empty while aliquota_iva is zero", code="00413")
        if self.aliquota_iva != 0 and self.natura is not None:
            validation.add_error(
                self._meta["natura"], "field has value while aliquota_iva is not zero", code="00414")


@export
class ScontoMaggiorazione(models.Model):
    tipo = fields.StringField(length=2, choices=("SC", "MG"))
    percentuale = fields.DecimalField(max_length=6, null=True)
    importo = fields.DecimalField(max_length=15, null=True)


@export
class DatiGeneraliDocumento(models.Model):
    tipo_documento = fields.StringField(length=4, choices=consts.TIPO_DOCUMENTO)
    divisa = fields.StringField()
    data = fields.DateField()
    numero = fields.StringField(max_length=20)
    dati_ritenuta = fields.ModelField(DatiRitenuta, null=True)
    dati_bollo = fields.ModelField(DatiBollo, null=True)
    dati_cassa_previdenziale = fields.ModelListField(DatiCassaPrevidenziale, null=True)
    sconto_maggiorazione = fields.ModelListField(ScontoMaggiorazione, null=True)
    importo_totale_documento = fields.DecimalField(max_length=15, null=True)
    arrotondamento = fields.DecimalField(max_length=15, null=True)
    causale = fields.ListField(fields.StringField(max_length=200), null=True)
    art73 = fields.StringField(length=2, choices=("SI",), null=True, xmltag="Art73")

    def validate_model(self, validation):
        super().validate_model(validation)

        has_dati_cassa_previdenziale_ritenuta = False
        for dcp in self.dati_cassa_previdenziale:
            if dcp.ritenuta == "SI":
                has_dati_cassa_previdenziale_ritenuta = True
                break

        if has_dati_cassa_previdenziale_ritenuta and not self.dati_ritenuta.has_value():
            validation.add_error(
                self._meta["ritenuta"], "field empty when dati_cassa_previdenziale.ritenuta is SI", code="00415")

        if self.numero is None or not re.search(r"\d", self.numero):
            validation.add_error(
                self._meta["numero"], "numero must contain at least one number", code="00425")


@export
class AltriDatiGestionali(models.Model):
    tipo_dato = fields.StringField(max_length=10)
    riferimento_testo = fields.StringField(max_length=60, null=True)
    riferimento_numero = fields.DecimalField(max_length=21, null=True)
    riferimento_data = fields.DateField(null=True)


@export
class CodiceArticolo(models.Model):
    codice_tipo = fields.StringField(max_length=35)
    codice_valore = fields.StringField(max_length=35)


@export
class DettaglioLinee(models.Model):
    numero_linea = fields.IntegerField(max_length=4)
    tipo_cessione_prestazione = fields.StringField(length=2, choices=("SC", "PR", "AB", "AC"), null=True)
    codice_articolo = fields.ModelListField(CodiceArticolo, null=True)
    descrizione = fields.StringField(max_length=1000)
    quantita = fields.DecimalField(max_length=21, decimals=2, null=True)
    unita_misura = fields.StringField(max_length=10, null=True)
    data_inizio_periodo = fields.DateField(null=True)
    data_fine_periodo = fields.DateField(null=True)
    prezzo_unitario = fields.DecimalField(max_length=21, decimals=3)
    sconto_maggiorazione = fields.ModelListField(ScontoMaggiorazione, null=True)
    prezzo_totale = fields.DecimalField(max_length=21)
    aliquota_iva = fields.DecimalField(xmltag="AliquotaIVA", max_length=6)
    ritenuta = fields.StringField(length=2, choices=("SI",), null=True)
    natura = fields.StringField(min_length=2, max_length=4, null=True, choices=consts.NATURA_IVA)
    riferimento_amministrazione = fields.StringField(max_length=20, null=True)
    altri_dati_gestionali = fields.ModelListField(AltriDatiGestionali, null=True)

    def validate_model(self, validation):
        super().validate_model(validation)
        if self.quantita is None and self.unita_misura is not None:
            validation.add_error(self._meta["quantita"], "field must be present when unita_misura is set")
        if self.quantita is not None and self.unita_misura is None:
            validation.add_error(self._meta["unita_misura"], "field must be present when quantita is set")
        if self.aliquota_iva == 0 and self.natura is None:
            validation.add_error(
                self._meta["natura"], "natura non presente a fronte di aliquota_iva pari a zero", code="00400")
        if self.aliquota_iva != 0 and self.natura is not None:
            validation.add_error(
                self._meta["natura"], "natura presente a fronte di aliquota_iva diversa da zero", code="00401")

    def autofill_prezzo_totale(self):
        """
        Compute prezzo_totale if it was not set
        """
        if self.prezzo_totale is not None:
            return
        if self.quantita is None:
            self.prezzo_totale = self.prezzo_unitario
        else:
            self.prezzo_totale = self.prezzo_unitario * self.quantita


@export
class DatiRiepilogo(models.Model):
    aliquota_iva = fields.DecimalField(xmltag="AliquotaIVA", max_length=6)
    natura = fields.StringField(min_length=2, max_length=4, null=True, choices=consts.NATURA_IVA)
    spese_accessorie = fields.DecimalField(max_length=15, null=True)
    arrotondamento = fields.DecimalField(max_length=21, null=True)
    # FIXME: Su questo valore il sistema effettua un controllo per verificare
    # la correttezza del calcolo; per i dettagli sull’algoritmo di calcolo si
    # rimanda al file Elenco controlli versione 1.4 presente sul sito
    # www.fatturapa.gov.it.
    imponibile_importo = fields.DecimalField(max_length=15)
    imposta = fields.DecimalField(max_length=15)
    esigibilita_iva = fields.StringField(xmltag="EsigibilitaIVA", length=1, choices=("I", "D", "S"), null=True)
    riferimento_normativo = fields.StringField(max_length=100, null=True)

    def validate_model(self, validation):
        super().validate_model(validation)
        if self.aliquota_iva == 0 and self.natura is None:
            validation.add_error(
                self._meta["natura"], "field is empty while aliquota_iva is zero", code="00429")
        if self.aliquota_iva != 0 and self.natura is not None:
            validation.add_error(
                self._meta["natura"], "field has value while aliquota_iva is not zero", code="00430")


@export
class DatiBeniServizi(models.Model):
    dettaglio_linee = fields.ModelListField(DettaglioLinee)
    dati_riepilogo = fields.ModelListField(DatiRiepilogo)

    def add_dettaglio_linee(self, **kw):
        """
        Convenience method to add entries to dettaglio_linee, autocomputing
        numero_linea and prezzo_totale when missing.

        prezzo_totale is just computed as prezzo_unitario * quantita.

        For anything more complicated, you need to compute prezzo_totale
        yourself add pass it explicitly, or better, extend this function and
        submit a pull request.
        """
        kw.setdefault("numero_linea", len(self.dettaglio_linee) + 1)
        d = DettaglioLinee(**kw)
        d.autofill_prezzo_totale()
        self.dettaglio_linee.append(d)

    def build_dati_riepilogo(self):
        """
        Convenience method to compute dati_riepilogo. It replaces existing
        values in dati_riepilogo.

        It only groups dettaglio_linee by aliquota, sums prezzo_totale to
        compute imponibile, and applies IVA.

        For anything more complicated, you need to compute dati_riepilogo
        yourself, or better, extend this function and submit a pull request.
        """
        from collections import defaultdict

        # Group by aliquota
        by_aliquota = defaultdict(list)
        for linea in self.dettaglio_linee:
            by_aliquota[(linea.aliquota_iva, linea.natura)].append(linea)

        self.dati_riepilogo = []
        for (aliquota, natura), linee in sorted(by_aliquota.items()):
            imponibile = sum(linea.prezzo_totale for linea in linee)
            imposta = imponibile * aliquota / 100
            self.dati_riepilogo.append(
                    DatiRiepilogo(
                        aliquota_iva=aliquota, imponibile_importo=imponibile,
                        imposta=imposta, esigibilita_iva="I", natura=natura))


@export
class DatiDocumentiCorrelati(models.Model):
    riferimento_numero_linea = fields.ListField(fields.IntegerField(max_length=4), null=True)
    id_documento = fields.StringField(max_length=20)
    data = fields.DateField(null=True)
    num_item = fields.StringField(max_length=20, null=True)
    codice_commessa_convenzione = fields.StringField(max_length=100, null=True)
    codice_cup = fields.StringField(max_length=15, xmltag="CodiceCUP", null=True)
    codice_cig = fields.StringField(max_length=15, xmltag="CodiceCIG", null=True)


@export
class DatiOrdineAcquisto(DatiDocumentiCorrelati):
    pass


@export
class DatiContratto(DatiDocumentiCorrelati):
    pass


@export
class DatiConvenzione(DatiDocumentiCorrelati):
    pass


@export
class DatiRicezione(DatiDocumentiCorrelati):
    pass


@export
class DatiFattureCollegate(DatiDocumentiCorrelati):
    pass


@export
class DatiAnagraficiVettore(models.Model):
    id_fiscale_iva = IdFiscaleIVA
    codice_fiscale = fields.StringField(min_length=11, max_length=16, null=True)
    anagrafica = Anagrafica
    numero_licenza_guida = fields.StringField(max_length=20, null=True)


@export
class IndirizzoResa(IndirizzoType):
    pass


@export
class DatiTrasporto(models.Model):
    dati_anagrafici_vettore = fields.ModelField(DatiAnagraficiVettore, null=True)
    mezzo_trasporto = fields.StringField(max_length=80, null=True)
    causale_trasporto = fields.StringField(max_length=100, null=True)
    numero_colli = fields.IntegerField(max_length=4, null=True)
    descrizione = fields.StringField(max_length=100, null=True)
    unita_misura_peso = fields.StringField(max_length=10, null=True)
    peso_lordo = fields.DecimalField(max_length=7, null=True)
    peso_netto = fields.DecimalField(max_length=7, null=True)
    data_ora_ritiro = fields.DateTimeField(null=True)
    data_inizio_trasporto = fields.DateField(null=True)
    tipo_resa = fields.StringField(length=3, null=True)
    indirizzo_resa = fields.ModelField(IndirizzoResa, null=True)
    data_ora_consegna = fields.DateTimeField(null=True)


@export
class DatiDDT(models.Model):
    __xmltag__ = "DatiDDT"
    numero_ddt = fields.StringField(max_length=20, xmltag="NumeroDDT")
    data_ddt = fields.DateField(xmltag="DataDDT")
    riferimento_numero_linea = fields.ListField(fields.IntegerField(max_length=4), null=True)


@export
class FatturaPrincipale(models.Model):
    numero_fattura_principale = fields.StringField(max_length=20)
    data_fattura_principale = fields.DateField()


@export
class DatiGenerali(models.Model):
    dati_generali_documento = DatiGeneraliDocumento
    dati_ordine_acquisto = fields.ModelListField(DatiOrdineAcquisto, null=True)
    dati_contratto = fields.ModelListField(DatiContratto, null=True)
    dati_convenzione = fields.ModelListField(DatiConvenzione, null=True)
    dati_ricezione = fields.ModelListField(DatiRicezione, null=True)
    dati_fatture_collegate = fields.ModelListField(DatiFattureCollegate, null=True)
    # dati_sal =
    dati_ddt = fields.ModelListField(DatiDDT, null=True)
    dati_trasporto = fields.ModelField(DatiTrasporto, null=True)
    fattura_principale = fields.ModelField(FatturaPrincipale, null=True)

    def validate_model(self, validation):
        super().validate_model(validation)
        dfc_dates = [x.data for x in self.dati_fatture_collegate if x.data is not None]
        if dfc_dates and self.dati_generali_documento.data < min(dfc_dates):
            validation.add_error(
                (self.dati_fatture_collegate._meta["data"],
                 self.dati_generali_documento._meta["data"]),
                "dati_generali_documento.data is earlier than dati_fatture_collegate.data",
                code="00418")


@export
class DettaglioPagamento(models.Model):
    beneficiario = fields.StringField(max_length=200, null=True)
    modalita_pagamento = fields.StringField(length=4, choices=consts.MODALITA_PAGAMENTO)
    data_riferimento_termini_pagamento = fields.DateField(null=True)
    giorni_termini_pagamento = fields.IntegerField(max_length=3, null=True)
    data_scadenza_pagamento = fields.DateField(null=True)
    importo_pagamento = fields.DecimalField(max_length=15)
    cod_ufficio_postale = fields.StringField(max_length=20, null=True)
    cognome_quietanzante = fields.StringField(max_length=60, null=True)
    nome_quietanzante = fields.StringField(max_length=60, null=True)
    cf_quietanzante = fields.StringField(max_length=16, null=True, xmltag="CFQuietanzante")
    titolo_quietanzante = fields.StringField(min_length=2, max_length=10, null=True)
    istituto_finanziario = fields.StringField(max_length=80, null=True)
    iban = fields.StringField(min_length=15, max_length=34, null=True, xmltag="IBAN")
    abi = fields.StringField(length=5, null=True, xmltag="ABI")
    cab = fields.StringField(length=5, null=True, xmltag="CAB")
    bic = fields.StringField(min_length=8, max_length=11, null=True, xmltag="BIC")
    sconto_pagamento_anticipato = fields.DecimalField(max_length=15, null=True)
    data_limite_pagamento_anticipato = fields.DateField(null=True)
    penalita_pagamenti_ritardati = fields.DecimalField(max_length=15, null=True)
    data_decorrenza_penale = fields.DateField(null=True)
    codice_pagamento = fields.StringField(max_length=60, null=True)


@export
class DatiPagamento(models.Model):
    condizioni_pagamento = fields.StringField(length=4, choices=("TP01", "TP02", "TP03"))
    dettaglio_pagamento = fields.ModelListField(DettaglioPagamento)


@export
class Allegati(models.Model):
    nome_attachment = fields.StringField(max_length=60)
    algoritmo_compressione = fields.StringField(max_length=10, null=True)
    formato_attachment = fields.StringField(max_length=10, null=True)
    descrizione_attachment = fields.StringField(max_length=100, null=True)
    attachment = fields.Base64BinaryField()


@export
class DatiVeicoli(models.Model):
    data = fields.DateField()
    totale_percorso = fields.StringField(max_length=15)


@export
class FatturaElettronicaBody(models.Model):
    dati_generali = DatiGenerali
    dati_beni_servizi = DatiBeniServizi
    dati_veicoli = models.ModelField(DatiVeicoli, null=True)
    dati_pagamento = fields.ModelListField(DatiPagamento, null=True)
    allegati = fields.ModelListField(Allegati, null=True)

    def build_importo_totale_documento(self):
        """
        Convenience method to compute
        dati_generali.dati_generali_documento.importo_totale_documento.
        It replaces an existing value in importo_totale_documento.

        It only adds imponibile_importo and imposta values from
        dati_beni_servizi.dati_riepilogo.

        For anything more complicated, you need to compute
        importo_totale_documento yourself, or better, extend this function and
        submit a pull request.
        """
        totale = sum(r.imponibile_importo + r.imposta for r in self.dati_beni_servizi.dati_riepilogo)
        self.dati_generali.dati_generali_documento.importo_totale_documento = totale

    def validate_model(self, validation):
        super().validate_model(validation)

        has_ritenute = False
        has_aliquote_iva = False
        for dl in self.dati_beni_servizi.dettaglio_linee:
            if dl.ritenuta == "SI":
                has_ritenute = True
            if dl.aliquota_iva is not None:
                has_aliquote_iva = True

        if has_ritenute and not self.dati_generali.dati_generali_documento.dati_ritenuta.has_value():
            validation.add_error(
                self.dati_generali.dati_generali_documento._meta["dati_ritenuta"],
                "field empty while at least one of dati_beni_servizi.dettaglio_linee.ritenuta is SI",
                code="00411",
            )

        for dcp in self.dati_generali.dati_generali_documento.dati_cassa_previdenziale:
            if dcp.aliquota_iva is not None:
                has_aliquote_iva = True

        if not self.dati_beni_servizi.dati_riepilogo and has_aliquote_iva:
            validation.add_error(
                    self.dati_beni_servizi._meta["dati_riepilogo"],
                    "dati_riepilogo is empty while there is at least an aliquota_iva"
                    " in dettaglio_linee or dati_cassa_previdenziale",
                    code="00419")


@export
class Fattura(models.Model):
    __xmlns__ = NS
    __xmltag__ = "FatturaElettronica"

    fattura_elettronica_header = FatturaElettronicaHeader
    fattura_elettronica_body = fields.ModelListField(FatturaElettronicaBody, min_num=1)
    signature = fields.NotImplementedField(null=True, xmlns=NS_SIG)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.fattura_elettronica_header.dati_trasmissione.formato_trasmissione = self.get_versione()

    def get_versione(self):
        return None

    def get_xmlattrs(self):
        return {"versione": self.get_versione()}

    def validate_model(self, validation):
        super().validate_model(validation)
        if self.get_versione() != self.fattura_elettronica_header.dati_trasmissione.formato_trasmissione:
            validation.add_error(
                    self.fattura_elettronica_header.dati_trasmissione._meta["formato_trasmissione"],
                    "formato_trasmissione should be {}".format(self.get_versione()),
                    code="00428")

    def to_xml(self, builder):
        with builder.element(self.get_xmltag(), **self.get_xmlattrs()) as b:
            with b.override_default_namespace(None) as b1:
                for name, field in self._meta.items():
                    field.to_xml(b1, getattr(self, name))

    def build_etree(self, lxml=False):
        """
        Build and return an ElementTree with the fattura in XML format
        """
        self.fattura_elettronica_header.dati_trasmissione.formato_trasmissione = self.get_versione()
        if lxml:
            from a38.builder import LXMLBuilder
            builder = LXMLBuilder()
        else:
            from a38.builder import Builder
            builder = Builder()
        builder.default_namespace = NS
        self.to_xml(builder)
        return builder.get_tree()

    def from_etree(self, el):
        versione = el.attrib.get("versione", None)
        if versione is None:
            raise RuntimeError("root element {} misses attribute 'versione'".format(el.tag))

        if versione != self.get_versione():
            raise RuntimeError("root element versione is {} instead of {}".format(versione, self.get_versione()))

        return super().from_etree(el)


@export
class FatturaPrivati12(Fattura):
    """
    Fattura privati 1.2
    """
    def get_versione(self):
        return "FPR12"


@export
class FatturaPA12(Fattura):
    """
    Fattura PA 1.2
    """
    def get_versione(self):
        return "FPA12"


@export
def auto_from_etree(root: ET.ElementTree) -> Union[Fattura, FatturaElettronicaSemplificata]:
    """
    Instantiate a Fattura or FatturaElettronicaSemplificata from a parsed XML
    """
    from .fattura_semplificata import NS10, FatturaElettronicaSemplificata
    tagname_ordinaria = "{{{}}}FatturaElettronica".format(NS)
    tagname_semplificata = "{{{}}}FatturaElettronicaSemplificata".format(NS10)

    versione = root.attrib.get("versione", None)

    if root.tag == tagname_ordinaria:
        if versione is None:
            raise RuntimeError("root element {} misses attribute 'versione'".format(root.tag))
        if versione == "FPR12":
            res = FatturaPrivati12()
        elif versione == "FPA12":
            res = FatturaPA12()
        else:
            raise RuntimeError("unsupported versione {}".format(versione))
    elif root.tag == tagname_semplificata:
        if versione is None:
            raise RuntimeError("root element {} misses attribute 'versione'".format(root.tag))
        if versione == "FSM10":
            res = FatturaElettronicaSemplificata()
        else:
            raise RuntimeError("unsupported versione {}".format(versione))
    else:
        raise RuntimeError("Root element {} is neither {} nor {}".format(
            root.tag, tagname_ordinaria, tagname_semplificata))

    res.from_etree(root)
    return res


@export
def auto_from_dict(data: Dict[str, Any]) -> Union[Fattura, FatturaElettronicaSemplificata]:
    """
    Given the equivalent of a Fattura.to_jsonable() data structure,
    reconstruct the fattura
    """
    from .fattura_semplificata import FatturaElettronicaSemplificata
    try:
        formato = data["fattura_elettronica_header"]["dati_trasmissione"]["formato_trasmissione"]
    except KeyError:
        raise RuntimeError("fattura_elettronica_header.dati_trasmissione.formato_trasmissione not found in input")

    if formato == "FPR12":
        return FatturaPrivati12(**data)
    elif formato == "FPA12":
        return FatturaPA12(**data)
    elif formato == "FSM10":
        return FatturaElettronicaSemplificata(**data)
    else:
        raise RuntimeError(f"Unsupported formato_trasmissione: {formato!r}")



from __future__ import annotations

from collections import defaultdict
from typing import Any, Dict, Optional, Tuple

from .fields import Field, ModelField
from .validation import Validation


class ModelBase:
    __slots__ = ()

    def __init__(self):
        pass

    @classmethod
    def get_xmltag(cls) -> str:
        xmltag = getattr(cls, "__xmltag__", None)
        if xmltag is None:
            xmltag = cls.__name__

        xmlns = getattr(cls, "__xmlns__", None)
        if xmlns:
            return "{" + xmlns + "}" + xmltag
        else:
            return xmltag

    def get_xmlattrs(self) -> Dict[str, str]:
        return {}


class ModelMetaclass(type):
    def __new__(cls, name, bases, dct):
        _meta = {}

        # Add fields from subclasses
        for b in bases:
            if not issubclass(b, ModelBase):
                continue
            b_meta = getattr(b, "_meta", None)
            if b_meta is None:
                continue
            _meta.update(b_meta)

        # Add fields from the class itself
        slots = []
        for field_name, val in list(dct.items()):
            if isinstance(val, Field):
                # Store its description in the Model _meta
                _meta[field_name] = val
                val.set_name(field_name)
            elif isinstance(val, type) and issubclass(val, ModelBase):
                # Store its description in the Model _meta
                val = ModelField(val)
                _meta[field_name] = val
                val.set_name(field_name)
            else:
                # Leave untouched
                continue

            # Remove field_name from class variables
            del dct[field_name]
            # Add it as a slot in the instance
            slots.append(field_name)

        dct["__slots__"] = slots
        res = super().__new__(cls, name, bases, dct)
        res._meta = _meta
        return res


class Model(ModelBase, metaclass=ModelMetaclass):
    """
    Declarative description of a data structure that can be validated and
    serialized to XML.
    """
    __slots__ = ()

    def __init__(self, *args, **kw):
        super().__init__()
        for name, value in zip(self._meta.keys(), args):
            kw[name] = value

        for name, field in self._meta.items():
            value = kw.pop(name, None)
            if value is None:
                value = field.get_construct_default()
            else:
                value = field.clean_value(value)
            setattr(self, name, value)

    def update(self, *args, **kw):
        """
        Set multiple values in the model.

        Arguments are treated in the same way as in the constructor. Any field
        not mentioned is left untouched.
        """
        for name, value in zip(self._meta.keys(), args):
            setattr(self, name, value)

        for name, value in kw.items():
            setattr(self, name, value)

    def has_value(self):
        for name, field in self._meta.items():
            if field.has_value(getattr(self, name)):
                return True
        return False

    @classmethod
    def clean_value(cls, value: Any) -> Optional["Model"]:
        """
        Create a model from the given value.

        Always make a copy even if value is already of the right class, to
        prevent mutability issues.
        """
        if value is None:
            return None
        if isinstance(value, dict):
            return cls(**value)
        elif isinstance(value, ModelBase):
            kw = {}
            for name, field in cls._meta.items():
                kw[name] = getattr(value, name, None)
            return cls(**kw)
        else:
            raise TypeError(f"{cls.__name__}: {value!r} is {type(value).__name__}"
                            " instead of a Model or dict instance")

    def validate_fields(self, validation: Validation):
        for name, field in self._meta.items():
            field.validate(validation, getattr(self, name))

    def validate_model(self, validation: Validation):
        pass

    def validate(self, validation: Validation):
        self.validate_fields(validation)
        self.validate_model(validation)

    def to_jsonable(self):
        res = {}
        for name, field in self._meta.items():
            value = field.to_jsonable(getattr(self, name))
            if value is not None:
                res[name] = value
        return res

    def to_python(self, **kw) -> str:
        args = []
        for name, field in self._meta.items():
            value = getattr(self, name)
            if not field.has_value(value):
                continue
            args.append(name + "=" + field.to_python(value, **kw))
        namespace = kw.get("namespace")
        if namespace is None:
            constructor = self.__class__.__module__ + "." + self.__class__.__qualname__
        elif namespace is False:
            constructor = self.__class__.__qualname__
        else:
            constructor = namespace + "." + self.__class__.__qualname__
        return "{}({})".format(constructor, ", ".join(args))

    def to_xml(self, builder):
        with builder.element(self.get_xmltag(), **self.get_xmlattrs()) as b:
            for name, field in self._meta.items():
                field.to_xml(b, getattr(self, name))

    def __setattr__(self, key: str, value: any):
        field = self._meta.get(key, None)
        if field is not None:
            value = field.clean_value(value)
        super().__setattr__(key, value)

    def _to_tuple(self) -> Tuple[Any]:
        return tuple(getattr(self, name) for name in self._meta.keys())

    def __eq__(self, other):
        other = self.clean_value(other)
        has_self = self.has_value()
        has_other = other is not None and other.has_value()
        if not has_self and not has_other:
            return True
        if has_self != has_other:
            return False
        return self._to_tuple() == other._to_tuple()

    def __ne__(self, other):
        other = self.clean_value(other)
        has_self = self.has_value()
        has_other = other is not None and other.has_value()
        if not has_self and not has_other:
            return False
        if has_self != has_other:
            return True
        return self._to_tuple() != other._to_tuple()

    def __lt__(self, other):
        other = self.clean_value(other)
        has_self = self.has_value()
        has_other = other is not None and other.has_value()
        if not has_self and not has_other:
            return False
        if has_self and not has_other:
            return False
        if not has_self and has_other:
            return True
        return self._to_tuple() < other._to_tuple()

    def __gt__(self, other):
        other = self.clean_value(other)
        has_self = self.has_value()
        has_other = other is not None and other.has_value()
        if not has_self and not has_other:
            return False
        if has_self and not has_other:
            return True
        if not has_self and has_other:
            return False
        return self._to_tuple() > other._to_tuple()

    def __le__(self, other):
        other = self.clean_value(other)
        has_self = self.has_value()
        has_other = other is not None and other.has_value()
        if not has_self and not has_other:
            return True
        if has_self and not has_other:
            return False
        if not has_self and has_other:
            return True
        return self._to_tuple() <= other._to_tuple()

    def __ge__(self, other):
        other = self.clean_value(other)
        has_self = self.has_value()
        has_other = other is not None and other.has_value()
        if not has_self and not has_other:
            return True
        if has_self and not has_other:
            return True
        if not has_self and has_other:
            return False
        return self._to_tuple() >= other._to_tuple()

    def __str__(self):
        vals = []
        for name, field in self._meta.items():
            vals.append(name + "=" + field.to_str(getattr(self, name)))
        return "{}({})".format(self.__class__.__name__, ", ".join(vals))

    def __repr__(self):
        vals = []
        for name, field in self._meta.items():
            vals.append(name + "=" + field.to_str(getattr(self, name)))
        return "{}({})".format(self.__class__.__name__, ", ".join(vals))

    def from_etree(self, el):
        if el.tag != self.get_xmltag():
            raise RuntimeError("element is {} instead of {}".format(el.tag, self.get_xmltag()))

        tag_map = {field.get_xmltag(): (name, field) for name, field in self._meta.items()}

        # Group values by tag
        by_name = defaultdict(list)
        for child in el:
            try:
                name, field = tag_map[child.tag]
            except KeyError:
                raise RuntimeError("found unexpected element {} in {}".format(child.tag, el.tag))

            by_name[name].append(child)

        for name, elements in by_name.items():
            field = self._meta[name]
            if field.multivalue:
                setattr(self, name, field.from_etree(elements))
            elif len(elements) != 1:
                raise RuntimeError(
                        "found {} {} elements in {} instead of just 1".format(
                            len(elements), child.tag, el.tag))
            else:
                setattr(self, name, field.from_etree(elements[0]))

    def diff(self, diff, other):
        has_self = self.has_value()
        has_other = other.has_value()
        if not has_self and not has_other:
            return
        if has_self != has_other:
            diff.add(None, self, other)
            return
        for name, field in self._meta.items():
            first = getattr(self, name)
            second = getattr(other, name)
            field.diff(diff, first, second)



from typing import Any, List, Optional

from . import fields
from .traversal import Annotation, Traversal


class Difference(Annotation):
    def __init__(self, prefix: Optional[str], field: "fields.Field", first: Any, second: Any):
        super().__init__(prefix, field)
        self.first = first
        self.second = second

    def __str__(self):
        return "{}: first: {}, second: {}".format(
                self.qualified_field,
                self.field.to_str(self.first),
                self.field.to_str(self.second))


class MissingOne(Difference):
    def __str__(self):
        if self.first is None:
            return "{}: first is not set".format(self.qualified_field)
        else:
            return "{}: second is not set".format(self.qualified_field)


class ExtraItems(Difference):
    def __str__(self):
        if len(self.first) > len(self.second):
            diff = len(self.first) - len(self.second)
            longer = "first"
        else:
            diff = len(self.second) - len(self.first)
            longer = "second"

        if diff == 1:
            return "{}: {} has 1 extra element".format(self.qualified_field, longer)
        else:
            return "{}: {} has {} extra elements".format(self.qualified_field, longer, diff)


class Diff(Traversal):
    def __init__(self, prefix: Optional[str] = None, differences: Optional[List[Difference]] = None):
        super().__init__(prefix)
        self.differences: List[Difference]
        if differences is None:
            self.differences = []
        else:
            self.differences = differences

    def with_prefix(self, prefix: str):
        return Diff(prefix, self.differences)

    def add_different(self, field: "fields.Field", first: Any, second: Any):
        self.differences.append(Difference(self.prefix, field, first, second))

    def add_only_first(self, field: "fields.Field", first: Any):
        self.differences.append(MissingOne(self.prefix, field, first, None))

    def add_only_second(self, field: "fields.Field", second: Any):
        self.differences.append(MissingOne(self.prefix, field, None, second))

    def add_different_length(self, field: "fields.Field", first: Any, second: Any):
        self.differences.append(ExtraItems(self.prefix, field, first, second))



import os
import subprocess
import tempfile
from typing import Optional

try:
    import lxml.etree
    HAVE_LXML = True
except ModuleNotFoundError:
    HAVE_LXML = False


if HAVE_LXML:
    class XSLTTransform:
        def __init__(self, xslt):
            parsed_xslt = lxml.etree.parse(xslt)
            self.xslt = lxml.etree.XSLT(parsed_xslt)

        def __call__(self, f):
            """
            Return the ElementTree for f rendered as HTML
            """
            tree = f.build_etree(lxml=True)
            return self.xslt(tree)

        def _requires_enable_local_file_access(self, wkhtmltopdf: str):
            """
            Check if we need to pass --enable-local-file-access to wkhtmltopdf.

            See https://github.com/Truelite/python-a38/issues/6 for details
            """
            # We need to specifically use --extended-help, because --help does
            # not always document --enable-local-file-access
            verifyLocalAccessToFileOption = subprocess.run(
                    [wkhtmltopdf, "--extended-help"], stdin=subprocess.DEVNULL, text=True, capture_output=True)
            return "--enable-local-file-access" in verifyLocalAccessToFileOption.stdout

        def to_pdf(self, wkhtmltopdf: str, f, output_file: Optional[str] = None):
            """
            Render a fattura to PDF using the given wkhtmltopdf command.

            Returns None if output_file is given, or the binary PDF data if not
            """
            if output_file is None:
                output_file = "-"
            html = self(f)

            # TODO: pass html data as stdin, using '-' as input for
            #       wkhtmltopdf: that currently removes the requirement for
            #       --enable-local-file-access
            with tempfile.NamedTemporaryFile("wb", suffix=".html", delete=False) as fd:
                html.write(fd)
                tempFilename = fd.name

            try:
                cmdLine = [wkhtmltopdf, tempFilename, output_file]
                if self._requires_enable_local_file_access(wkhtmltopdf):
                    cmdLine.insert(1, "--enable-local-file-access")

                res = subprocess.run(cmdLine, stdin=subprocess.DEVNULL, capture_output=True)

                if res.returncode != 0:
                    raise RuntimeError(
                            "{0} exited with error {1}: stderr: {2!r}".format(
                                wkhtmltopdf, res.returncode, res.stderr))

                if output_file == "-":
                    return res.stdout
                else:
                    return None
            finally:
                os.remove(tempFilename)



from __future__ import annotations

import io
import json
import logging
import os
import subprocess
import tempfile

try:
    from defusedxml import ElementTree as ET
except ModuleNotFoundError:
    import xml.etree.ElementTree as ET

from typing import (Any, BinaryIO, Dict, List, Optional, Sequence, TextIO,
                    Type, Union)

try:
    import ruamel.yaml
    yaml = None
except ModuleNotFoundError:
    ruamel = None
    try:
        import yaml
    except ModuleNotFoundError:
        yaml = None

from . import crypto
from .fattura import auto_from_dict, auto_from_etree
from .models import Model

log = logging.getLogger("codec")

if ruamel is not None:
    def _load_yaml(fd: TextIO):
        yaml_loader = ruamel.yaml.YAML(typ="safe", pure=True)
        return yaml_loader.load(fd)

    def _write_yaml(data: Dict[str, Any], file: TextIO):
        yaml = ruamel.yaml.YAML(typ="safe")
        yaml.default_flow_style = False
        yaml.allow_unicode = True
        yaml.explicit_start = True
        yaml.dump(data, file)
elif yaml is not None:
    def _load_yaml(fd: TextIO):
        return yaml.load(fd, Loader=yaml.CLoader)

    def _write_yaml(data: Dict[str, Any], file: TextIO):
        yaml.dump(
            data, stream=file, default_flow_style=False,
            allow_unicode=True, explicit_start=True, Dumper=yaml.CDumper)
else:
    def _load_yaml(fd: TextIO):
        raise NotImplementedError("loading YAML requires ruamel.yaml or PyYAML to be installed")

    def _write_yaml(data: Dict[str, Any], file: TextIO):
        raise NotImplementedError("writing YAML requires ruamel.yaml or PyYAML to be installed")


class Codec:
    """
    Base class for format-specific reading and writing of fatture
    """
    # If True, file objects are expected to be open in binary mode
    binary = False

    def load(
            self,
            pathname: str,
            model: Optional[Type[Model]]) -> Model:
        """
        Load a fattura from a file.

        If model is provided it will be used for loading, otherwise the Model
        type will be autodetected
        """
        raise NotImplementedError(f"{self.__class__.__name__}.load is not implemented")

    def write_file(self, f: Model, file: Union[TextIO, BinaryIO]):
        """
        Write a fattura to the given file deescriptor.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.write_file is not implemented")

    def save(self, f: Model, pathname: str):
        """
        Write a fattura to the given file
        """
        with open(pathname, "wb" if self.binary else "wt") as fd:
            self.write_file(f, fd)

    def interactive_edit(self, f: Model) -> Optional[Model]:
        """
        Edit the given model in an interactive editor, using the format of this
        codec
        """
        with io.StringIO() as orig:
            self.write_file(f, orig)
            return self.edit_buffer(orig.getvalue(), model=f.__class__)

    def edit_buffer(self, buf: str, model: Optional[Type[Model]] = None) -> Optional[Model]:
        """
        Open an editor on buf and return the edited fattura.

        Return None if editing did not change the contents.
        """
        editor = os.environ.get("EDITOR", "sensible-editor")

        current = buf
        error = None

        while True:
            with tempfile.NamedTemporaryFile(
                    mode="wt",
                    suffix=f".{self.EXTENSIONS[0]}") as tf:
                # Write out the current buffer
                tf.write(current)
                if error is not None:
                    tf.write(f"# ERROR: {error}")
                    error = None
                tf.flush()

                # Run the editor on it
                subprocess.run([editor, tf.name], check=True)

                # Reopen by name in case the editor did not write on the same
                # inode
                with open(tf.name, "rt") as fd:
                    lines = []
                    for line in fd:
                        if line.startswith("# ERROR: "):
                            continue
                        lines.append(line)
                    edited = "".join(lines)

                if edited == current:
                    return None

                try:
                    return self.load(tf.name, model=model)
                except Exception as e:
                    log.error("%s: cannot load edited file: %s", tf.name, e)
                    error = str(e)


class P7M(Codec):
    """
    P7M codec, that only supports loading
    """
    EXTENSIONS = ("p7m",)

    def load(
            self,
            pathname: str,
            model: Optional[Type[Model]] = None) -> Model:
        p7m = crypto.P7M(pathname)
        return p7m.get_fattura()


class JSON(Codec):
    """
    JSON codec.

    `indent` represents the JSON structure indentation, and can be None to
    output everything in a single line.

    `end` is a string that gets appended to the JSON structure.
    """
    EXTENSIONS = ("json",)

    def __init__(self, indent: Optional[int] = 1, end="\n"):
        self.indent = indent
        self.end = end

    def load(
            self,
            pathname: str,
            model: Optional[Type[Model]] = None) -> Model:
        with open(pathname, "rt") as fd:
            data = json.load(fd)
        if model:
            return model(**data)
        else:
            return auto_from_dict(data)

    def write_file(self, f: Model, file: TextIO):
        json.dump(f.to_jsonable(), file, indent=self.indent)
        if self.end is not None:
            file.write(self.end)


class YAML(Codec):
    """
    YAML codec
    """
    EXTENSIONS = ("yaml", "yml")

    def load(
            self,
            pathname: str,
            model: Optional[Type[Model]] = None) -> Model:
        with open(pathname, "rt") as fd:
            data = _load_yaml(fd)
        if model:
            return model(**data)
        else:
            return auto_from_dict(data)

    def write_file(self, f: Model, file: TextIO):
        _write_yaml(f.to_jsonable(), file)


class Python(Codec):
    """
    Python codec.

    `namespace` defines what namespace is used to refer to `a38` models. `None`
    means use a default, `False` means not to use a namespace, a string defines
    which namespace to use.

    `unformatted` can be set to True to skip code formatting.

    The code will be written with just the expression to build the fattura.

    The code assumes `import datetime` and `from decimal import Decimal`.

    If loadable is True, the file is written as a Python source that
    creates a `fattura` variable with the fattura, with all the imports that
    are needed. This generates a python file that can be loaded with load().

    Note that loading Python fatture executes arbitrary Python code!
    """
    EXTENSIONS = ("py",)

    def __init__(
            self, namespace: Union[None, bool, str] = "a38",
            unformatted: bool = False,
            loadable: bool = False):
        self.namespace = namespace
        self.unformatted = unformatted
        self.loadable = loadable

    def load(
            self,
            pathname: str,
            model: Optional[Type[Model]] = None) -> Model:
        with open(pathname, "rt") as fd:
            code = compile(fd.read(), pathname, 'exec')

        loc = {}
        exec(code, {}, loc)
        return loc["fattura"]

    def write_file(self, f: Model, file: TextIO):
        code = f.to_python(namespace=self.namespace)

        if not self.unformatted:
            try:
                from yapf.yapflib import yapf_api
            except ModuleNotFoundError:
                pass
            else:
                code, changed = yapf_api.FormatCode(code)

        if self.loadable:
            print("import datetime", file=file)
            print("from decimal import Decimal", file=file)
            if self.namespace:
                print("import", self.namespace, file=file)
            elif self.namespace is False:
                print("from a38.fattura import *", file=file)
            else:
                print("import a38", file=file)
            print(file=file)
            print("fattura = ", file=file, end="")
        print(code, file=file)


class XML(Codec):
    """
    XML codec
    """
    EXTENSIONS = ("xml",)

    binary = True

    def load(
            self,
            pathname: str,
            model: Optional[Type[Model]] = None) -> Model:
        tree = ET.parse(pathname)
        return auto_from_etree(tree.getroot())

    def write_file(self, f: Model, file: BinaryIO):
        tree = f.build_etree()
        tree.write(file, encoding="utf-8", xml_declaration=True)
        file.write(b"\n")


class Codecs:
    """
    A collection of codecs
    """
    ALL_CODECS = (XML, P7M, JSON, YAML, Python)

    def __init__(
            self,
            include: Optional[Sequence[Type[Codec]]] = None,
            exclude: Optional[Sequence[Type[Codec]]] = (Python,)):
        """
        if `include` is not None, only codecs in that list are used.

        If `exclude` is not None, all codecs are used except the given one.

        If neither `include` nor `exclude` are None, all codecs are used.

        By default, `exclude` is not None but it is set to exclude Python.
        """
        self.codecs: List[Type[Codec]]

        if include is not None and exclude is not None:
            raise ValueError("include and exclude cannot both be set")
        elif include is not None:
            self.codecs = list(include)
        elif exclude is not None:
            self.codecs = [c for c in self.ALL_CODECS if c not in exclude]
        else:
            self.codecs = list(self.ALL_CODECS)

    def codec_from_filename(self, pathname: str) -> Type[Codec]:
        """
        Infer a Codec class from the extension of the file at `pathname`.
        """
        ext = pathname.rsplit(".", 1)[1].lower()

        for c in self.codecs:
            if ext in c.EXTENSIONS:
                return c
